import logging

import numpy as np

import finesse.detectors as fd

from finesse.plotting.tools import add_colorbar
from finesse.utilities.units import get_SI_value


LOGGER = logging.getLogger(__name__)


def bode(f, *Y, axs=None, return_axes=True, figsize=(6, 6), **kwargs):
    """Create a Bode plot for a complex array.

    Examples
    --------
    >>> axs = bode(f, CLG, label='CLG')
    >>> bode(f, CHR, axs=axs, label='CHR')
    """

    import matplotlib.pyplot as plt

    if axs is None:
        fig, axs = plt.subplots(2, 1, sharex=True, figsize=figsize)

    axs[0].set_ylabel("Amplitude [dB]")
    axs[1].set_xlabel("Frequency [Hz]")
    axs[1].set_ylabel("Phase [Deg]")

    for y in Y:
        mag = 20.0 * np.log10(abs(y))
        phase = np.unwrap(np.arctan2(y.imag, y.real)) * 180.0 / np.pi
        axs[0].semilogx(f, mag, **kwargs)
        axs[1].semilogx(f, phase, **kwargs)

    if "label" in kwargs:
        axs[0].legend()
        axs[1].legend()

    if return_axes:
        return axs


def ws_phase_space(
    W,
    S,
    OL,
    cmap="bone",
    levels=None,
    wscale="mm",
    sscale="m",
    contour_kwargs=None,
    clabel_kwargs=None,
    show=True,
    fig=None,
    ax=None,
):
    """Plots the overlap contours for WS phase space data.

    The return values of :func:`.ws_overlap_grid` correspond to the first three
    arguments of this function.

    Parameters
    ----------
    W : :class:`numpy.ndarray`
        The W space (as a 2D grid).

    S : :class:`numpy.ndarray`
        The S space (as a 2D grid).

    OL : :class:`numpy.ndarray`
        The overlap as a function of the WS phase space (as a 2D grid).

    cmap : str or colormap, optional; default: "bone"
        A matplotlib colormap, or its name.

    levels : list, optional; default: None
        List of contour levels to pass to contour plotting explicitly.

    wscale : str, optional; default: "mm"
        Units for W-axis (i.e. beam size units).

    sscale : str, optional; default: "m"
        Reciprocal units for S-axis (i.e. defocus units).

    contour_kwargs : dict, optional
        Dictionary of keyword arguments to pass to matplotlib contour function. If not
        specified then the following defaults are used:

        - "colors": "k"
        - "linestyles": "--"
        - "linewidths": 0.5

    clabel_kwargs : dict, optional
        Dictionary of keyword arguments to pass to matplotlib clabel function. If not
        specified then the following defaults are used:

        - "colors": same as `contour_kwargs`
        - "inline": True

    show : bool, optional; default: True
        Whether to show the figure immediately.

    fig : :class:`~matplotlib.figure.Figure`, optional, default: None
        The figure object to use. If not specified a new figure will be drawn.

    ax : :class:`~matplotlib.axes.Axes`, optional, default: None
        The axes to use. If not specified the first pair will be used, or created.
        Ignored if `fig` is None.

    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        Handle to the matplotlib Figure.

    ax : :class:`~matplotlib.axes.Axes`
        Handle to the matplotlib Axis.

    See Also
    --------
    :class:`~finesse.gaussian.ws_overlap_grid`
    """
    import matplotlib.pyplot as plt

    if fig is None:
        fig = plt.figure()
        if ax is not None:
            LOGGER.warn(
                "Axes have been specified without figure. Axes will be ignored."
            )
            ax = None

    if ax is None:
        if len(fig.axes) == 0:
            ax = fig.add_subplot()
        else:
            LOGGER.warn("Axes not specified, using first pair.")
            ax = fig.axes[0]

    if contour_kwargs is None:
        contour_kwargs = {}
    if clabel_kwargs is None:
        clabel_kwargs = {}

    # Set some sensible defaults for the overlap contour line properties
    contour_kwargs.setdefault("colors", "k")
    contour_kwargs.setdefault("linestyles", "--")
    contour_kwargs.setdefault("linewidths", 0.5)

    # And some defaults for the overlap contour label properties
    clabel_kwargs.setdefault("inline", True)
    clabel_kwargs.setdefault("colors", contour_kwargs["colors"])

    CS = ax.contourf(W, S, OL, cmap=plt.cm.get_cmap(cmap), levels=levels)
    CS2 = ax.contour(W, S, OL, levels=CS.levels, **contour_kwargs)
    ax.clabel(CS2, **clabel_kwargs)

    if len(wscale) == 2:
        wsv = get_SI_value(wscale[0])
        ax.set_xticklabels(f"{(x / wsv):.2f}" for x in ax.get_xticks())
    if len(sscale) == 2:
        ssv = get_SI_value(sscale[0])
        ax.set_yticklabels(f"{(y / ssv):.2f}" for y in ax.get_yticks())

    ax.set_xlabel(f"Gaussian mode size $W$ [{wscale}]")
    ax.set_ylabel(f"Gaussian defocus $S$ [1/{sscale}]")

    if show:
        plt.show()

    return fig, ax


class Plotter:
    """Handler for plotting outputs from a simulation."""

    import matplotlib.pyplot as plt
    import matplotlib.colors as colors
    import matplotlib.animation as animation

    def __init__(self, solution):
        self.out = solution
        # figure size scaling and layout
        self.scale = 1
        self.tight_layout = True

        # animation attributes
        self.repeat = True
        self.repeat_delay = None
        self.interval = 50
        self.blit = True

    def __get_detectors(self, detectors=None):
        if detectors is None:
            detectors = self.out.detectors
            assert detectors is not None

        elif isinstance(detectors, type) or (
            hasattr(detectors, "__getitem__")
            and all(isinstance(d, type) for d in detectors)
        ):
            if hasattr(detectors, "__getitem__"):
                det_types = detectors
            else:
                det_types = [detectors]

            detectors = list(filter(lambda x: type(x) in det_types, self.out.detectors))

        elif isinstance(detectors, str) or isinstance(detectors, fd.Detector):
            detectors = [detectors]

        elif hasattr(detectors, "__getitem__") and all(
            isinstance(d, (str, fd.Detector)) for d in detectors
        ):
            pass

        else:
            LOGGER.warning("Invalid 'detectors' argument passed to plotting.")
            detectors = []

        for i, detector in enumerate(detectors):
            if isinstance(detector, str):
                try:
                    b = filter(lambda x: x.name == detector, self.out.detectors)
                    element = tuple(b)[0]
                except IndexError:
                    element = None

                msg = ""
                if element is None:
                    msg = "No object named %s found in the model."
                if not isinstance(element, fd.Detector):
                    msg = "Object %s is not a detector!"

                if msg:
                    LOGGER.warning(msg, detector)
                    continue

                detectors[i] = element

        return detectors

    def __do_scaling_and_layout(self, figures):
        done = set()
        for figs_t in figures.values():
            if isinstance(figs_t, dict):
                figs = list(figs_t.values())
                if figs and isinstance(figs[0], list):
                    figs = figs[0]
            elif isinstance(figs_t, list):
                figs = figs_t
            else:
                figs = [figs_t]

            for fig in figs:
                if fig in done:
                    continue

                fig.set_size_inches(self.scale * fig.get_size_inches())

                if self.tight_layout:
                    fig.tight_layout()

                done.add(fig)

    @staticmethod
    def parameter_axis_label(param, axis="x", ax=None):
        if ax is not None:
            if axis == "x":
                func = ax.set_xlabel
            else:
                func = ax.set_ylabel
        else:
            if axis == "x":
                func = Plotter.plt.xlabel
            else:
                func = Plotter.plt.ylabel

        label = f"{param.component.name} {param.name}"
        if param.units:
            label += f" [{param.units}]"

        func(label)

    @staticmethod
    def output_axis_label(obj, ax=None):
        if ax is not None:
            func = ax.set_ylabel
        else:
            func = Plotter.plt.ylabel

        if obj.label is None:
            label = ""
            LOGGER.warning(
                "Detector %s has no label, unable to set output "
                "axis label for this detector.",
                obj.name,
            )

        label = f"{obj.label}"

        if obj.unit:
            label += f" [{obj.unit}]"

        func(label)

    @staticmethod
    def choose_plot_func(logx, logy, magnitude_axis=None):
        if magnitude_axis is not None:
            obj = magnitude_axis
        else:
            obj = Plotter.plt

        if logx and logy:
            return obj.loglog
        if logx:
            return obj.semilogx
        if logy:
            return obj.semilogy

        return obj.plot

    @staticmethod
    def select_detector_cmap(cmaps, det):
        det_type = type(det)

        if isinstance(cmaps, dict):
            if det.name in cmaps:
                cmap = cmaps[det.name]
            elif det_type in cmaps:
                cmap = cmaps[det_type]
            else:
                LOGGER.info(
                    "No entry for %s or %s in specified cmap dictionary, "
                    "using default colormap",
                    det.name,
                    det_type,
                )
                cmap = Plotter.plt.get_cmap()
        else:
            cmap = cmaps

        return cmap

    @staticmethod
    def make_text_handle(
        text_former, units, x, y, index=0, color="white", fig=None, ax=None
    ):
        units_txt = f"\n[{units}]" if units else ""

        if ax is None:
            if fig is None:
                raise RuntimeError()

            txt_func = Plotter.plt.text
            transform = fig.axes[0].transAxes
        else:
            txt_func = ax.text
            transform = ax.transAxes

        return txt_func(
            x,
            y,
            text_former(index) + units_txt,
            ha="center",
            va="center",
            transform=transform,
            color=color,
        )

    @staticmethod
    def make_images(
        fig,
        extent,
        z,
        cmap,
        det,
        log,
        anim_axis,
        aspect="auto",
        cbar_label=None,
        transpose=False,
    ):
        x, p = anim_axis

        norm = None
        if log:
            norm = Plotter.colors.LogNorm(z.min(), z.max())

        if transpose:
            z0 = z[0].T
        else:
            z0 = z[0]

        initial_im = Plotter.plt.imshow(
            z0, norm=norm, extent=extent, cmap=cmap, aspect=aspect, animated=True,
        )
        if cbar_label is None:
            cbar_label = f"{det.label}"
            if det.unit:
                cbar_label += f" [{det.unit}]"
            cbar_label += f" ({det.name})"

        add_colorbar(initial_im, label=cbar_label)

        def form_txt(idx):
            return f"{p.component.name} {p.name} = {x[idx]:.3g}"

        txt_handle = Plotter.make_text_handle(form_txt, p.units, x=0.8, y=0.9, fig=fig)

        images = [[initial_im, txt_handle]]
        for i in np.arange(1, x.size):
            if transpose:
                zi = z[i].T
            else:
                zi = z[i]

            im = Plotter.plt.imshow(
                zi, norm=norm, extent=extent, cmap=cmap, aspect=aspect, animated=True,
            )
            txt_handle = Plotter.make_text_handle(
                form_txt, p.units, x=0.8, y=0.9, index=i, fig=fig,
            )

            images.append([im, txt_handle])

        return images

    @staticmethod
    def make_amp_phase_images(
        ax1,
        ax2,
        extent,
        z1,
        z2,
        cmap,
        degrees,
        det,
        log,
        anim_axis,
        aspect="auto",
        transpose=False,
    ):
        x, p = anim_axis

        norm = None
        if log:
            norm = Plotter.colors.LogNorm(z1.min(), z1.max())

        if transpose:
            z10 = z1[0].T
            z20 = z2[0].T
        else:
            z10 = z1[0]
            z20 = z2[0]

        initial_amp_im = ax1.imshow(
            z10, norm=norm, extent=extent, cmap=cmap, aspect=aspect, animated=True,
        )
        add_colorbar(
            initial_amp_im, label=r"Amplitude [$\sqrt{{W}}$" + f" ({det.name})"
        )

        initial_phase_im = ax2.imshow(
            z20, extent=extent, cmap=cmap, aspect=aspect, animated=True,
        )
        add_colorbar(
            initial_phase_im,
            label=f"Phase [{'deg' if degrees else 'rad'}] ({det.name})",
        )

        def form_txt(idx):
            return f"{p.component.name} {p.name} = {x[idx]:.3g}"

        amp_txt_handle = Plotter.make_text_handle(
            form_txt, p.units, x=0.8, y=0.9, ax=ax1
        )
        phase_txt_handle = Plotter.make_text_handle(
            form_txt, p.units, x=0.8, y=0.9, ax=ax2
        )

        images = [[initial_amp_im, initial_phase_im, amp_txt_handle, phase_txt_handle]]

        for i in np.arange(1, x.size):
            if transpose:
                z1i = z1[i].T
                z2i = z2[i].T
            else:
                z1i = z1[i]
                z2i = z2[i]

            amp_im = ax1.imshow(
                z1i, norm=norm, extent=extent, cmap=cmap, aspect=aspect, animated=True,
            )
            phase_im = ax2.imshow(
                z2i, extent=extent, cmap=cmap, aspect=aspect, animated=True,
            )

            amp_txt_handle = Plotter.make_text_handle(
                form_txt, p.units, x=0.8, y=0.9, index=i, ax=ax1,
            )

            phase_txt_handle = Plotter.make_text_handle(
                form_txt, p.units, x=0.8, y=0.9, index=i, ax=ax2,
            )

            images.append([amp_im, phase_im, amp_txt_handle, phase_txt_handle])

        return images

    @staticmethod
    def _set_fig_at_detname(figures: dict, dets, fig):
        if not hasattr(dets, "__getitem__"):
            dets = [dets]

        figures.update(dict.fromkeys([det.name for det in dets], fig))

    def __handle_beam_property_plotting(
        self, bp_detector_map, cmaps, figures, animations, logx, logy, log, degrees
    ):
        if bp_detector_map:
            sub_figs = figures[fd.BeamPropertyDetector] = {}

        for det_prop, dets in bp_detector_map.items():
            if det_prop == fd.BeamProperty.Q:
                continue

            if self.out.axes == 1:
                self.__plot_1D(det_prop, dets, logx, logy, degrees, sub_figs, figures)
            elif self.out.axes == 2:
                self.__plot_2D(det_prop, dets, log, degrees, cmaps, sub_figs, figures)
            elif self.out.axes == 3:
                self.__plot_2D_animated(
                    det_prop, dets, log, degrees, cmaps, sub_figs, animations, figures
                )

    def __handle_cavity_property_plotting(
        self, cp_detector_map, cmaps, figures, animations, logx, logy, log, degrees
    ):
        if cp_detector_map:
            sub_figs = figures[fd.CavityPropertyDetector] = {}

        for det_prop, dets in cp_detector_map.items():
            if det_prop == fd.CavityProperty.EIGENMODE:
                continue

            if self.out.axes == 1:
                self.__plot_1D(det_prop, dets, logx, logy, degrees, sub_figs, figures)
            elif self.out.axes == 2:
                self.__plot_2D(det_prop, dets, log, degrees, cmaps, sub_figs, figures)
            elif self.out.axes == 3:
                self.__plot_2D_animated(
                    det_prop, dets, log, degrees, cmaps, sub_figs, animations, figures
                )

    def __handle_detector_plotting(
        self, detector_type_map, cmaps, figures, animations, logx, logy, log, degrees,
    ):
        for det_type, dets in detector_type_map.items():
            if det_type == fd.CCD:
                self.__plot_CCDs(dets, log, cmaps, figures, animations)
            elif det_type == fd.FieldCamera:
                self.__plot_field_cameras(
                    dets, log, degrees, cmaps, figures, animations
                )
            elif det_type == fd.CCDScanLine:
                self.__plot_ccd_scan_lines(dets, log, cmaps, figures, animations)
            elif det_type == fd.FieldScanLine:
                self.__plot_field_scan_lines(
                    dets, log, degrees, cmaps, figures, animations
                )
            else:
                if not self.out.axes:
                    LOGGER.warning(
                        "No x-axes have been defined, unable to plot the "
                        "outputs of any detectors of type %s",
                        det_type,
                    )
                    continue

                if self.out.axes == 1:
                    self.__plot_1D(det_type, dets, logx, logy, degrees, figures)
                elif self.out.axes == 2:
                    self.__plot_2D(det_type, dets, log, degrees, cmaps, figures)
                elif self.out.axes == 3:
                    self.__plot_2D_animated(
                        det_type, dets, log, degrees, cmaps, figures, animations
                    )
                else:
                    LOGGER.error(
                        "Unable to produce plots of %d-dimensional data", self.out.axes
                    )

    def __plot_1D(self, det_type, dets, logx, logy, degrees, figures, allfigs=None):
        fig = figures.get(det_type)
        if fig is None:
            fig = Plotter.plt.figure()
        else:
            Plotter.plt.figure(fig.number)

        mag_and_phase = (
            allfigs is None
            and issubclass(det_type, fd.Detector)
            and any(det.dtype == np.complex128 for det in dets)
        )
        if mag_and_phase:
            mag_ax = fig.add_subplot(211)
            phase_ax = fig.add_subplot(212, sharex=mag_ax)
        else:
            mag_ax = None

        plot_func = Plotter.choose_plot_func(logx, logy, mag_ax)

        for det in dets:
            data = self.out[det]

            if mag_and_phase:
                amplitude = np.abs(data)
                plot_func(self.out.x1, amplitude, label=det.name + " (abs)")
                mag_ax.set_ylabel(r"Amplitude [$\sqrt{W}$]")
                mag_ax.legend()

                phase = np.angle(data, degrees)

                phase_ax.plot(self.out.x1, phase, label=det.name + " (phase)")
                phase_ax.set_ylabel(f"Phase [{'deg' if degrees else 'rad'}]")
            else:
                plot_func(self.out.x1, data, label=det.name)

        if not mag_and_phase and dets:
            Plotter.output_axis_label(dets[0])
            Plotter.plt.legend()

        Plotter.parameter_axis_label(self.out.p1)

        figures[det_type] = fig

        if allfigs is None:
            Plotter._set_fig_at_detname(figures, dets, fig)
        else:
            Plotter._set_fig_at_detname(allfigs, dets, fig)

    def __plot_2D(self, det_type, dets, log, degrees, cmaps, figures, allfigs=None):
        x1 = self.out.x1
        x2 = self.out.x2
        extent = [x1.min(), x1.max(), x2.min(), x2.max()]

        for det in dets:
            fig = Plotter.plt.figure()

            data = self.out[det].T
            cmap = Plotter.select_detector_cmap(cmaps, det)

            mag_and_phase = det.dtype == np.complex128
            if mag_and_phase:
                mag_ax = fig.add_subplot(211)
                phase_ax = fig.add_subplot(212)

                amplitude = np.abs(data)

                norm = None
                if log:
                    norm = Plotter.colors.LogNorm(amplitude.min(), amplitude.max())

                mag_im = mag_ax.imshow(
                    amplitude, extent=extent, norm=norm, cmap=cmap, aspect="auto",
                )
                add_colorbar(
                    mag_im, label=r"Amplitude [$\sqrt{{W}}$]" + f" ({det.name})"
                )

                phase = np.angle(data, degrees)
                phase_im = phase_ax.imshow(
                    phase, extent=extent, cmap=cmap, aspect="auto",
                )
                add_colorbar(
                    phase_im,
                    label=f"Phase [{'deg' if degrees else 'rad'}] ({det.name})",
                )
                for ax in (mag_ax, phase_ax):
                    Plotter.parameter_axis_label(self.out.p1, ax=ax)
                    Plotter.parameter_axis_label(self.out.p2, axis="y", ax=ax)

            else:
                norm = None
                if log:
                    norm = Plotter.colors.LogNorm(data.min(), data.max())

                im = Plotter.plt.imshow(
                    data, extent=extent, norm=norm, cmap=cmap, aspect="auto",
                )
                add_colorbar(im, label=f"{det.label} [{det.unit}] ({det.name})")

                Plotter.parameter_axis_label(self.out.p1)
                Plotter.parameter_axis_label(self.out.p2, axis="y")

            if det_type in figures:
                figures[det_type].append(fig)
            else:
                figures[det_type] = [fig]

            if allfigs is None:
                Plotter._set_fig_at_detname(figures, det, fig)
            else:
                Plotter._set_fig_at_detname(allfigs, det, fig)

    def __plot_2D_animated(
        self, det_type, dets, log, degrees, cmaps, figures, animations, allfigs=None,
    ):
        x1 = self.out.x1
        x2 = self.out.x2
        extent = [x1.min(), x1.max(), x2.min(), x2.max()]

        for det in dets:
            fig = Plotter.plt.figure()

            data = self.out[det]
            cmap = Plotter.select_detector_cmap(cmaps, det)

            mag_and_phase = det.dtype == np.complex128
            if mag_and_phase:
                mag_ax = fig.add_subplot(211)
                phase_ax = fig.add_subplot(212)

                amplitude = np.abs(data)
                phase = np.angle(data, degrees)

                images = Plotter.make_amp_phase_images(
                    mag_ax,
                    phase_ax,
                    extent,
                    amplitude,
                    phase,
                    cmap,
                    degrees,
                    det,
                    log,
                    (self.out.x3, self.out.p3),
                    transpose=True,
                )

                for ax in (mag_ax, phase_ax):
                    Plotter.parameter_axis_label(self.out.p1, ax=ax)
                    Plotter.parameter_axis_label(self.out.p2, axis="y", ax=ax)
            else:
                images = Plotter.make_images(
                    fig,
                    extent,
                    data,
                    cmap,
                    det,
                    log,
                    (self.out.x3, self.out.p3),
                    transpose=True,
                )

                Plotter.parameter_axis_label(self.out.p1)
                Plotter.parameter_axis_label(self.out.p2, axis="y")

            animations[det.name] = Plotter.animation.ArtistAnimation(
                fig,
                images,
                interval=self.interval,
                repeat=self.repeat,
                repeat_delay=self.repeat_delay,
                blit=self.blit,
            )

            if det_type in figures:
                figures[det_type].append(fig)
            else:
                figures[det_type] = [fig]

            if allfigs is None:
                Plotter._set_fig_at_detname(figures, det, fig)
            else:
                Plotter._set_fig_at_detname(allfigs, det, fig)

    def __plot_CCDs(self, dets, log, cmaps, figures, animations):
        if self.out.axes > 1:
            LOGGER.warning("Skipping CCD plots as multiple axes have been defined")
            return

        for det in dets:
            fig = Plotter.plt.figure()

            data = self.out[det]
            norm = None
            if log:
                norm = Plotter.colors.LogNorm(data.min(), data.max())
            w0_scaled = det.w0_scaled

            cmap = Plotter.select_detector_cmap(cmaps, det)

            extent = [*det.xlim, *det.ylim]

            cb_label = r"Intensity [W m$^{-2}$ px]" + f" ({det.name})"

            if not self.out.axes:  # single image
                im = Plotter.plt.imshow(
                    data.T, norm=norm, extent=extent, cmap=cmap, aspect="equal",
                )
                add_colorbar(im, label=cb_label)

            else:
                images = Plotter.make_images(
                    fig,
                    extent,
                    data,
                    cmap,
                    det,
                    log,
                    (self.out.x1, self.out.p1),
                    aspect="equal",
                    cbar_label=cb_label,
                    transpose=True,
                )

                animations[det.name] = Plotter.animation.ArtistAnimation(
                    fig,
                    images,
                    interval=self.interval,
                    repeat=self.repeat,
                    repeat_delay=self.repeat_delay,
                    blit=self.blit,
                )

            Plotter.plt.xlabel("$x$ " + "[$w_0$]" if w0_scaled else "[m]")
            Plotter.plt.ylabel("$y$ " + "[$w_0$]" if w0_scaled else "[m]")

            if fd.CCD in figures:
                figures[fd.CCD].append(fig)
            else:
                figures[fd.CCD] = [fig]

            Plotter._set_fig_at_detname(figures, det, fig)

    def __plot_field_cameras(self, dets, log, degrees, cmaps, figures, animations):
        if self.out.axes > 1:
            LOGGER.warning(
                "Skipping ComplexCamera plots as multiple axes have been defined"
            )
            return

        for det in dets:
            fig = Plotter.plt.figure()

            data = self.out[det]
            norm = None
            if log:
                norm = Plotter.colors.LogNorm(data.min(), data.max())
            w0_scaled = det.w0_scaled

            cmap = Plotter.select_detector_cmap(cmaps, det)

            extent = [*det.xlim, *det.ylim]

            mag_ax = fig.add_subplot(211)
            phase_ax = fig.add_subplot(212)

            amplitude = np.abs(data)
            phase = np.angle(data, degrees)
            if not self.out.axes:  # single image
                mag_im = mag_ax.imshow(
                    amplitude.T, norm=norm, extent=extent, cmap=cmap, aspect="equal",
                )
                add_colorbar(mag_im, label=det.name + r" Amplitude [$\sqrt{W}$ px]")

                phase_im = phase_ax.imshow(
                    phase.T, extent=extent, cmap=cmap, aspect="equal",
                )
                add_colorbar(
                    phase_im, label=det.name + f" Phase [{'deg' if degrees else 'rad'}]"
                )

            else:
                images = Plotter.make_amp_phase_images(
                    mag_ax,
                    phase_ax,
                    extent,
                    amplitude,
                    phase,
                    cmap,
                    degrees,
                    det,
                    log,
                    (self.out.x1, self.out.p1),
                    aspect="equal",
                    transpose=True,
                )

                animations[det.name] = Plotter.animation.ArtistAnimation(
                    fig,
                    images,
                    interval=self.interval,
                    repeat=self.repeat,
                    repeat_delay=self.repeat_delay,
                    blit=self.blit,
                )

            for ax in (mag_ax, phase_ax):
                ax.set_xlabel("$x$ " + "[$w_0$]" if w0_scaled else "[m]")
                ax.set_ylabel("$y$ " + "[$w_0$]" if w0_scaled else "[m]")

            if fd.FieldCamera in figures:
                figures[fd.FieldCamera].append(fig)
            else:
                figures[fd.FieldCamera] = [fig]

            Plotter._set_fig_at_detname(figures, det, fig)

    def __plot_ccd_scan_lines(
        self, dets, log, cmap, figures, animations,
    ):
        if self.out.axes > 2:
            LOGGER.warning(
                "Skipping CCDScanLine plots as more than two axes have been defined"
            )
            return

        if not self.out.axes:
            fig = Plotter.plt.figure()

        for det in dets:
            if self.out.axes:
                fig = Plotter.plt.figure()

            data = self.out[det]
            if det.direction == "x":
                x = det.xdata
            else:
                x = det.ydata
            w0_scaled = det.w0_scaled

            if not self.out.axes:
                plot_func = Plotter.choose_plot_func(log, log)

                plot_func(x, data, label=det.name)
            else:
                extent = [x.min(), x.max(), self.out.x1.min(), self.out.x1.max()]

                if self.out.axes == 1:
                    norm = None
                    if log:
                        norm = Plotter.colors.LogNorm(data.min(), data.max())

                    im = Plotter.plt.imshow(
                        data.T, extent=extent, norm=norm, cmap=cmap, aspect="auto",
                    )
                    add_colorbar(im, label=f"{det.label} [{det.unit}] ({det.name})")
                else:
                    images = Plotter.make_images(
                        fig,
                        extent,
                        data,
                        cmap,
                        det,
                        log,
                        (self.out.x2, self.out.p2),
                        transpose=True,
                    )

                    animations[det.name] = Plotter.animation.ArtistAnimation(
                        fig,
                        images,
                        interval=self.interval,
                        repeat=self.repeat,
                        repeat_delay=self.repeat_delay,
                        blit=self.blit,
                    )

                Plotter.parameter_axis_label(self.out.p1, axis="y")

                if fd.FieldScanLine in figures:
                    figures[fd.FieldScanLine].append(fig)
                else:
                    figures[fd.FieldScanLine] = [fig]

                Plotter._set_fig_at_detname(figures, det, fig)

            Plotter.plt.xlabel(
                f"${det.direction}$ " + "[$w_0$]" if w0_scaled else "[m]"
            )

        if not self.out.axes:
            Plotter.output_axis_label(dets[0])
            Plotter.plt.legend()

            figures[fd.FieldScanLine] = fig
            Plotter._set_fig_at_detname(figures, dets, fig)

    def __plot_field_scan_lines(
        self, dets, log, degrees, cmap, figures, animations,
    ):
        if self.out.axes > 2:
            LOGGER.warning(
                "Skipping FieldScanLine plots as more than two axes have been defined"
            )
            return

        if not self.out.axes:
            fig = Plotter.plt.figure()

        for det in dets:
            if self.out.axes:
                fig = Plotter.plt.figure()

            mag_ax = fig.add_subplot(211)
            phase_ax = fig.add_subplot(212)

            if det.direction == "x":
                x = det.xdata
            else:
                x = det.ydata
            w0_scaled = det.w0_scaled

            amplitude = np.abs(self.out[det])
            phase = np.angle(self.out[det], degrees)

            if not self.out.axes:
                plot_func = Plotter.choose_plot_func(log, log, mag_ax)

                plot_func(x, amplitude, label=det.name)
                mag_ax.set_ylabel(r"Amplitude [$\sqrt{W}$ px]")
                mag_ax.legend()

                phase_ax.plot(x, phase, label=det.name)
                phase_ax.set_ylabel(f"Phase [{'deg' if degrees else 'rad'}]")
                phase_ax.legend()
            else:
                extent = [x.min(), x.max(), self.out.x1.min(), self.out.x1.max()]

                if self.out.axes == 1:
                    norm = None
                    if log:
                        norm = Plotter.colors.LogNorm(amplitude.min(), amplitude.max())

                    mag_im = mag_ax.imshow(
                        amplitude.T, norm=norm, extent=extent, cmap=cmap, aspect="auto",
                    )
                    add_colorbar(mag_im, label=r"Ampltide [$\sqrt{W}$ px]")

                    phase_im = phase_ax.imshow(
                        phase.T, extent=extent, cmap=cmap, aspect="auto",
                    )
                    add_colorbar(
                        phase_im, label=f"Phase [{'deg' if degrees else 'rad'}]"
                    )

                else:
                    images = Plotter.make_amp_phase_images(
                        mag_ax,
                        phase_ax,
                        extent,
                        amplitude,
                        phase,
                        cmap,
                        degrees,
                        det,
                        log,
                        (self.out.x2, self.out.p2),
                        aspect="equal",
                        transpose=True,
                    )

                    animations[det.name] = Plotter.animation.ArtistAnimation(
                        fig,
                        images,
                        interval=self.interval,
                        repeat=self.repeat,
                        repeat_delay=self.repeat_delay,
                        blit=self.blit,
                    )

                for ax in (mag_ax, phase_ax):
                    Plotter.parameter_axis_label(self.out.p1, axis="y", ax=ax)

                if fd.FieldScanLine in figures:
                    figures[fd.FieldScanLine].append(fig)
                else:
                    figures[fd.FieldScanLine] = [fig]

                Plotter._set_fig_at_detname(figures, det, fig)

            for ax in (mag_ax, phase_ax):
                ax.set_xlabel(f"${det.direction}$ " + "[$w_0$]" if w0_scaled else "[m]")

        if not self.out.axes:
            figures[fd.FieldScanLine] = fig
            Plotter._set_fig_at_detname(figures, dets, fig)

    def plot(
        self,
        detectors=None,
        log=False,
        logx=None,
        logy=None,
        degrees=True,
        cmap=None,
        show=True,
        separate=True,
        _test_fig_handles=None,
    ):
        r"""Plots the outputs from the specified `detectors` of a given solution `out`, or all
        detectors in the executed model if `detectors` is `None`.

        Detectors are sorted by their type and the outputs of each are plotted on their own figure
        accordingly - i.e. all amplitude detector outputs are plotted on one figure, all power
        detector outputs on another figure etc.

        .. note::

            It is recommended to use this function with :func:`finesse.plotting.tools.init`. This
            then means that all figures produced by this function will use matplotlib rcParams
            corresponding to the style selected.

            For example::

                import finesse
                finesse.plotting.init()

                model = finesse.parse(\"""
                l L0 P=1

                s s0 L0.p1 ITM.p1

                m ITM R=0.99 T=0.01 Rc=inf
                s CAV ITM.p2 ETM.p1 L=1
                m ETM R=0.99 T=0.01 Rc=10

                modes maxtem=4

                gauss L0.p1.o q=(-0.4+2j)
                cav FP ITM.p2.o ITM.p2.i

                xaxis L0.f (-100M) 100M 1000 lin

                ad ad00 ETM.p1.i f=&L0.f n=0 m=0
                ad ad02 ETM.p1.i f=&L0.f n=0 m=2

                pd C ETM.p1.i
                \""")
                model.run().plot(logy=True, figsize_scale=2)

            will produce two figures (one for the power-detector output and another for the
            amplitude detectors) which use rcParams from the `default` style-sheet. Using
            `figsize_scale` here then scales these figures whilst keeping the proportions
            defined in this style-sheet constant.

        .. rubric:: Multi-dimensional scan plotting behaviour

        If multiple parameters have been scanned in the underlying model associated
        with this solution object, then the form of the resulting plots produced here
        will depend on a number of options:

        - If two parameters have been scanned then all non-CCD detector ouputs will be plotted
            on separate image plot figures. All CCD plots will be ignored.
        - If a single parameter has been scanned and `index` is not specified then all CCD detector
            outputs will be plotted on separate animated figures. Or if `index` is specified, then
            all CCD detector outputs will be plotted on separate image plot figures *at the
            specified index of the scanned axis*.

        Parameters
        ----------
        detectors : sequence or str or type or :class:`.Detector`, optional
            An iterable (or singular) of strings (corresponding to detector names),
            :class:`.Detector` instances or detector types. Defaults to `None` so that all detector
            outputs are plotted.

        log : bool, optional
            Use log-log scale. Also applies to image plots so that colours are normalised
            on a log-scale between limits of image data. Defaults to `False`.

        logx : bool, optional
            Use log-scale on x-axis if appropriate, defaults to the
            value of `log`.

        logy : bool, optional
            Use log-scale on y-axis if appropriate, defaults to the
            value of `log`.

        degrees : bool, optional
            Plots angle and phase outputs in degrees, defaults to `True`.

        show : bool, optional
            Shows all the produced figures if true, defaults to `True`.

        separate : bool, optional
            Plots the output of different detector types on different
            axes if true, defaults to `True`.

        Returns
        -------
        figures : dict
            A dictionary mapping `type(detector)` and detector names to corresponding
            :class:`~matplotlib.figure.Figure` objects. Note that some keys, i.e. those of each
            `type(detector)` and the names of the detectors of that type, share the same values.

        animations : dict
            A dictionary of `detector.name : animation` mappings.
        """
        if logx is None:
            logx = log
        if logy is None:
            logy = log
        detectors = self.__get_detectors(detectors)

        detector_type_map = {}
        bp_detector_map = {}
        cp_detector_map = {}

        for detector in detectors:
            if isinstance(detector, fd.BeamPropertyDetector):
                if detector.detecting in bp_detector_map:
                    bp_detector_map[detector.detecting].append(detector)
                else:
                    bp_detector_map[detector.detecting] = [detector]
            elif isinstance(detector, fd.CavityPropertyDetector):
                if detector.detecting in cp_detector_map:
                    cp_detector_map[detector.detecting].append(detector)
                else:
                    cp_detector_map[detector.detecting] = [detector]
            else:
                if separate:
                    key = type(detector)
                else:
                    key = type(fd.Detector)
                if key in detector_type_map:
                    detector_type_map[key].append(detector)
                else:
                    detector_type_map[key] = [detector]

        if cmap is None:
            cmap = Plotter.plt.get_cmap()

        figures = _test_fig_handles or {}
        animations = {}

        self.__handle_beam_property_plotting(
            bp_detector_map, cmap, figures, animations, logx, logy, log, degrees,
        )
        self.__handle_cavity_property_plotting(
            cp_detector_map, cmap, figures, animations, logx, logy, log, degrees,
        )
        self.__handle_detector_plotting(
            detector_type_map, cmap, figures, animations, logx, logy, log, degrees,
        )

        self.__do_scaling_and_layout(figures)

        if show:
            Plotter.plt.show()

        if not animations:
            return figures

        return figures, animations



"""
The ``detectors`` sub-module contains various non-physical detectors
that can be used to probe to simulation at any point.

Listed below are all the sub-modules within ``detectors`` with brief
descriptions of what is contained in each.
"""

from finesse.detectors.general import (
    Detector,
    MaskedDetector,
    NoiseDetector,
    SymbolDetector,
)
from finesse.detectors.amplitude_detector import AmplitudeDetector
from finesse.detectors.bpdetector import BeamProperty, BeamPropertyDetector
from finesse.detectors.camera import (
    Camera,
    CCDCamera,
    CCD,
    CCDScanLine,
    CCDPixel,
    ComplexCamera,
    FieldCamera,
    FieldScanLine,
    FieldPixel,
)
from finesse.detectors.cavity_detector import CavityProperty, CavityPropertyDetector
from finesse.detectors.gouy import Gouy
from finesse.detectors.astigmatism_detector import AstigmatismDetector
from finesse.detectors.knmdetector import KnmDetector
from finesse.detectors.mismatch_detector import ModeMismatchDetector
from finesse.detectors.motion_detector import MotionDetector
from finesse.detectors.powerdetector import (
    PowerDetector,
    PowerDetectorDemod1,
    PowerDetectorDemod2,
    CustomPD,
    SplitPD,
)
from finesse.detectors.quantum_noise_detector import (
    QuantumNoiseDetector,
    QuantumNoiseDetectorDemod1,
    QuantumNoiseDetectorDemod2,
    QuantumShotNoiseDetector,
    QuantumShotNoiseDetectorDemod1,
    QuantumShotNoiseDetectorDemod2,
    GeneralQuantumNoiseDetector,
)


__all__ = (
    "Detector",
    "MaskedDetector",
    "NoiseDetector",
    "SymbolDetector",
    "AmplitudeDetector",
    "BeamProperty",
    "BeamPropertyDetector",
    "Camera",
    "CCDCamera",
    "CCD",
    "CCDScanLine",
    "CCDPixel",
    "ComplexCamera",
    "FieldCamera",
    "FieldScanLine",
    "FieldPixel",
    "CavityProperty",
    "CavityPropertyDetector",
    "Gouy",
    "AstigmatismDetector",
    "KnmDetector",
    "ModeMismatchDetector",
    "MotionDetector",
    "PowerDetector",
    "PowerDetectorDemod1",
    "PowerDetectorDemod2",
    "CustomPD",
    "SplitPD",
    "QuantumNoiseDetector",
    "QuantumNoiseDetectorDemod1",
    "QuantumNoiseDetectorDemod2",
    "QuantumShotNoiseDetector",
    "QuantumShotNoiseDetectorDemod1",
    "QuantumShotNoiseDetectorDemod2",
    "GeneralQuantumNoiseDetector",
)



"""Tests for model trace forest structure when cavities which overlap are included."""

from itertools import permutations

import pytest
from finesse import Model
import finesse.ligo
from finesse.tracing.ctracer import TraceTree
from ...util import assert_trace_trees_equal


@pytest.fixture()
def aligo_model():
    IFO = Model()
    IFO.parse(finesse.ligo.aligo_katscript)
    return IFO


@pytest.mark.parametrize("order", list(permutations(["PRX", "PRY"])))
@pytest.mark.parametrize("with_arm_cavs", (True, False))
def test_overlapping_prx_pry_forest(aligo_model: Model, order, with_arm_cavs):
    """Test that overlapping power recycling cavities in X and Y arms are structured
    correctly in the model trace forest."""
    IFO = aligo_model

    IFO.parse(
        """
    cav PRX PRM.p2 ITMX.p1.i
    cav PRY PRM.p2 ITMY.p1.i
    """
    )

    if with_arm_cavs:
        IFO.parse(
            """
        cav cavXARM ITMX.p2
        cav cavYARM ITMY.p2
        """
        )

    IFO.beam_trace(order=order)

    expect = [TraceTree.from_cavity(getattr(IFO, cav)) for cav in order]
    expect.reverse()

    # Arm cavities don't overlap with anything so will constitute
    # the first block of internal trace trees
    if with_arm_cavs:
        expect.insert(0, TraceTree.from_cavity(IFO.cavXARM))
        expect.insert(1, TraceTree.from_cavity(IFO.cavYARM))

    # Only really care about the internal trace trees, other
    # tests cover external forest structures
    for i in range(len(expect)):
        assert_trace_trees_equal(IFO.trace_forest.forest[i], expect[i])


@pytest.mark.parametrize("order", list(permutations(["SRX", "SRY"])))
@pytest.mark.parametrize("with_arm_cavs", (True, False))
def test_overlapping_srx_sry_forest(aligo_model: Model, order, with_arm_cavs):
    """Test that overlapping signal recycling cavities in X and Y arms are structured
    correctly in the model trace forest."""
    IFO = aligo_model

    IFO.parse(
        """
    cav SRX SRM.p1 ITMX.p1.i
    cav SRY SRM.p1 ITMY.p1.i
    """
    )

    if with_arm_cavs:
        IFO.parse(
            """
        cav cavXARM ITMX.p2
        cav cavYARM ITMY.p2
        """
        )

    IFO.beam_trace(order=order)

    expect = [TraceTree.from_cavity(getattr(IFO, cav)) for cav in order]
    expect.reverse()

    # Arm cavities don't overlap with anything so will constitute
    # the first block of internal trace trees
    if with_arm_cavs:
        expect.insert(0, TraceTree.from_cavity(IFO.cavXARM))
        expect.insert(1, TraceTree.from_cavity(IFO.cavYARM))

    # Only really care about the internal trace trees, other
    # tests cover external forest structures
    for i in range(len(expect)):
        assert_trace_trees_equal(IFO.trace_forest.forest[i], expect[i])


@pytest.mark.parametrize("order", list(permutations(["SRX", "SRY", "PRX", "PRY"])))
def test_overlapping_all_recycling_cavs_forest(aligo_model: Model, order):
    """Test that overlapping both recycling cavities in X and Y arms are structured
    correctly in the model trace forest."""
    IFO = aligo_model

    IFO.parse(
        """
    cav SRX SRM.p1 ITMX.p1.i
    cav SRY SRM.p1 ITMY.p1.i
    cav PRX PRM.p2 ITMX.p1.i
    cav PRY PRM.p2 ITMY.p1.i
    """
    )

    IFO.beam_trace(order=order)

    expect = [TraceTree.from_cavity(getattr(IFO, cav)) for cav in order]
    expect.reverse()

    # Only really care about the internal trace trees, other
    # tests cover external forest structures
    for i in range(len(expect)):
        assert_trace_trees_equal(IFO.trace_forest.forest[i], expect[i])



from functools import partial
import pytest


@pytest.fixture
def compiler(monkeypatch, compiler):
    monkeypatch.setattr(
        compiler, "compile", partial(compiler.compile, resolve=True, build=False)
    )
    return compiler



"""Kat file compiler analysis parsing tests."""

from finesse.script import parse
from finesse.analysis.actions import Xaxis, BeamTrace, ABCD
from ....diff import assert_actions_equivalent


def test_nested_analysis():
    """Test that a kat script with a two level nested analysis builds correctly."""
    model = parse(
        """
        mirror m1
        xaxis(
            m1.phi,
            lin,
            0,
            90,
            100,
            pre_step=beam_trace(
                name="My Beam Trace"
            ),
            post_step=abcd(
                "My ABCD"
            ),
            name="myxaxis"
        )
        """
    )
    # FIXME: the assert function below doesn't check much here - needs improved.
    assert_actions_equivalent(
        model.analysis,
        Xaxis(
            model.m1.phi,
            "lin",
            0,
            90,
            100,
            pre_step=BeamTrace(name="My Beam Trace"),
            post_step=ABCD("My ABCD"),
            name="myxaxis",
        ),
    )


def test_parallel():
    model = parse(
        """
        l L0 P=1
        s s0 L0.p1 m1.p1
        m m1 R=0 T=1
        pd P m1.p2.o

        parallel(
            xaxis(m1.R, lin, 0, 1, 10),
            xaxis(L0.P, lin, 10, 100, 20, name="myxaxis")
        )
        """
    )
    # First xaxis.
    assert (
        model.analysis.actions[0].parameter
        == model.analysis.actions[0].parameter1
        == model.m1.R
    )
    assert model.analysis.actions[0].mode == model.analysis.actions[0].mode1 == "lin"
    assert model.analysis.actions[0].start == model.analysis.actions[0].start1 == 0
    assert model.analysis.actions[0].stop == model.analysis.actions[0].stop1 == 1
    assert model.analysis.actions[0].steps == model.analysis.actions[0].steps1 == 10
    assert model.analysis.actions[0].name == "xaxis"

    # Second xaxis.
    assert (
        model.analysis.actions[1].parameter
        == model.analysis.actions[1].parameter1
        == model.L0.P
    )
    assert model.analysis.actions[1].mode == model.analysis.actions[1].mode1 == "lin"
    assert model.analysis.actions[1].start == model.analysis.actions[1].start1 == 10
    assert model.analysis.actions[1].stop == model.analysis.actions[1].stop1 == 100
    assert model.analysis.actions[1].steps == model.analysis.actions[1].steps1 == 20
    assert model.analysis.actions[1].name == "myxaxis"

    # FIXME: test pre_step, etc. too - see #93


def test_expressions_as_action_arguments():
    """Test that action arguments can be expressions.

    See #130.
    """
    model = parse(
        """
        var f_m 1e5

        l pump P=1e3 f=-&f_m
        s s1 pump.p1 m1.p1
        m m1 T=5.9e-4 L=0
        s scav m1.p2 m2.p1 L=5
        m m2 T=1e-6 L=0

        ad aor m1.p1.o f=&fsig.f-&f_m
        ad aot m2.p2.o f=&fsig.f-&f_m

        sgen sig1 pump.amp.i 1 0

        fsig(10)
        xaxis(
            parameter=fsig.f,
            mode=lin,
            start=-5000+&f_m,
            stop=5000+&f_m,
            steps=300
        )
        """
    )

    assert float(model.analysis.start) == -5000 + 1e5
    assert float(model.analysis.stop) == 5000 + 1e5



import pytest
from finesse.script.exceptions import KatScriptError


def test_lambda_default(model):
    assert model.lambda0 == 1064e-9


@pytest.mark.parametrize("lambda0", (1, 3.141, 10))
def test_lambda(model, lambda0):
    model.parse(f"lambda({lambda0})")
    assert float(model.lambda0) == lambda0


@pytest.mark.xfail(reason="lambda0 validation not implemented yet")
@pytest.mark.parametrize("lambda0", (0, 0.0, -1, -3.141, -10, "'hi'", 1 + 2j))
def test_invalid_lambda(model, lambda0):
    with pytest.raises(KatScriptError):
        model.parse(f"lambda({lambda0})")



import pytest
from finesse.script.containers import KatScript, KatElement
from ..util import NAME, STRING, SPACE
from ..data import STRINGS


@pytest.mark.parametrize("string,_a,_b", STRINGS)
def test_string(parser, string, _a, _b):
    assert parser.parse(f"fake_element el1 {string}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[STRING(1, 18, string)],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )



"""Parse function tests."""

import pytest
from finesse.script import parse, parse_file
from ....diff import assert_models_equivalent
from ...util import dedent_multiline


@pytest.fixture
def model_kat_script():
    """Some uneventful kat model."""
    return dedent_multiline(
        """
        laser l0 P=1
        space s1 l0.p1 itm.p1
        mirror itm R=0.99 T=0.01
        space scav itm.p2 etm.p1 L=1
        mirror etm R=&itm.R T=&itm.T phi=&itm.phi
        pd pdcav itm.p1.o
        """
    )


@pytest.fixture
def model_via_string(model_kat_script):
    return parse(model_kat_script)


@pytest.fixture
def model_via_file(tmp_path, model_kat_script):
    """Test that parsing a string with kat script is the same as via a file."""
    tmpfile = tmp_path / "script.kat"
    with tmpfile.open(mode="w") as fp:
        fp.write(model_kat_script)
    with tmpfile.open(mode="r") as fp:
        return parse_file(fp)


def test_parse_function_is_same_as_model_parse_method(
    model_via_string, model, model_kat_script
):
    """Test that parsing directly with :meth:`.parse` results in the same model as
    :meth:`.Model.parse`."""
    model.parse(model_kat_script)
    assert_models_equivalent(model_via_string, model)


def test_parse_is_same_as_parse_file(tmp_path, model_via_string, model_via_file):
    """Test that parsing a string with kat script is the same as via a file."""
    assert_models_equivalent(model_via_string, model_via_file)



"""Kat script specification.

This defines supported kat script syntax and maps it to Finesse Python classes via
adapters.
"""

import abc
from difflib import get_close_matches
import logging
from collections import ChainMap
from .. import components, detectors, locks, symbols
from ..model import Model
from ..components import mechanical, electronics
from ..components.ligo import suspensions as ligo
from ..analysis import actions
from ..analysis import noise
from .adapter import ElementAdapter, CommandAdapter, AnalysisAdapter, GetterProxy


LOGGER = logging.getLogger(__name__)


class _GuassGetterProxy(GetterProxy):
    def __call__(self, gauss):
        # Use the :attr:`.Gauss._specified_params` attribute to dump the `kwargs`
        # signature argument.
        return [], {"name": gauss.name, "node": gauss.node, **gauss._specified_params}


class _FsigGetterProxy(GetterProxy):
    def __call__(self, model):
        if model.fsig.f.value is None:
            return

        return [model.fsig.f], {}


def _set_fsig(model, f):
    """Signal input frequency.

    Parameters
    ----------
    f : float or :class:`.Symbol`
        The frequency.
    """
    model.fsig.f = f


class _LambdaGetterProxy(GetterProxy):
    def __call__(self, model):
        return [model.lambda0], {}


def _set_lambda0(model, lambda0):
    """Reference wavelength.

    Parameters
    ----------
    lambda0 : float
        The reference wavelength.
    """
    model.lambda0 = lambda0


class _ModesGetterProxy(GetterProxy):
    def __call__(self, model):
        modes = model.modes_setting

        # Filter out empty values.
        modes = {key: value for key, value in modes.items() if value is not None}

        if not modes:
            return

        return [], modes


def _set_link(model, *args, **kwargs):
    model.link(*args, **kwargs)


class _IntrixGetterProxy(GetterProxy):
    def __call__(self, model):
        if not model.input_matrix_dc:
            return

        # FIXME: implement proper getter
        raise NotImplementedError("intrix dumping not yet implemented")


def _set_intrix(model, *args, **kwargs):
    """Set input matrix."""
    assert not kwargs

    DOF = args[0]
    factors, readouts = args[1::2], args[2::2]
    if len(factors) != len(readouts):
        raise Exception("must specify 'factor, readout' pairs")
    for factor, readout in zip(factors, readouts):
        model.input_matrix_dc[DOF, readout] = factor


class _TEMGetterProxy(GetterProxy):
    def __call__(self, model):
        """(args, kwargs) tuples for each defined TEM mode."""
        tems = []

        for laser in model.get_elements_of_type(components.Laser):
            for (n, m), (factor, phase) in laser.non_default_power_coeffs.items():
                tems.append(
                    ([laser], {"n": n, "m": m, "factor": factor, "phase": phase})
                )

        if not tems:
            # Don't generate anything.
            return

        return tems


def _set_tem(model, laser, *args, **kwargs):
    """Set laser TEM."""
    laser.tem(*args, **kwargs)


class BaseSpec(metaclass=abc.ABCMeta):
    """Empty language specification."""

    _SUPPORTED_CONSTANTS = {}
    _SUPPORTED_KEYWORDS = set()
    _SUPPORTED_UNARY_OPERATORS = {}
    _SUPPORTED_BINARY_OPERATORS = {}
    _SUPPORTED_EXPRESSION_FUNCTIONS = {}
    _DEFAULT_ELEMENTS = []
    _DEFAULT_COMMANDS = []
    _DEFAULT_ANALYSES = []

    def __init__(self):
        # Modifiable specifications. These are dynamically supported by the parser.
        self.elements = {}
        self.commands = {}
        self.analyses = {}

        # Fixed specifications. These are not modifiable by the user.
        self.constants = self._SUPPORTED_CONSTANTS
        self.keywords = self._SUPPORTED_KEYWORDS
        self.unary_operators = self._SUPPORTED_UNARY_OPERATORS
        self.binary_operators = self._SUPPORTED_BINARY_OPERATORS
        self.expression_functions = self._SUPPORTED_EXPRESSION_FUNCTIONS

        # Add support for the default directives.
        for elementdata in self._DEFAULT_ELEMENTS:
            self.register_element(*elementdata)
        for commanddata in self._DEFAULT_COMMANDS:
            self.register_command(*commanddata)
        for analysisdata in self._DEFAULT_ANALYSES:
            self.register_analysis(*analysisdata)

    @property
    def directives(self):
        """All top level parser directives.

        :getter: Returns a mapping of top level parser directive aliases to
                 :class:`adapters <.BaseAdapter>`.
        :type: :class:`~collections.ChainMap`
        """
        # ChainMap yields in LIFO order so key order becomes elements, then commands,
        # then analyses. This order is relied upon by :func:`.syntax`.
        return ChainMap(self.analyses, self.commands, self.elements)

    @property
    def reserved_names(self):
        """All reserved names.

        This is primarily useful for tests.

        :getter: Returns the names reserved in the parser as special production types.
        :type: :class:`list`
        """
        return list(self.keywords) + list(self.constants)

    def _register_adapter(self, ptype, mapping, aliases, kwargs=None, overwrite=False):
        if kwargs is None:
            kwargs = {}

        adapter = ptype(aliases, **kwargs)

        for alias in adapter.aliases:
            if alias in mapping:
                if overwrite:
                    LOGGER.info(f"Overwriting existing '{alias}' with {adapter}")
                else:
                    raise KeyError(
                        f"'{alias}' from {adapter} already exists (provided by "
                        f"{mapping[alias]}). If you intend to overwrite the existing "
                        f"definition, set overwrite=True."
                    )

            mapping[alias] = adapter

    def register_element(self, *args, **kwargs):
        """Add parser and generator support for a model element such as a component or
        detector.

        Other Parameters
        ----------------
        aliases : str or sequence
            The element alias(es).

        kwargs : mapping, optional
            Keyword arguments to pass to the adapter constructor; defaults to None.

        overwrite : bool, optional
            Overwrite elements with the same aliases, if present. Defaults to False.
        """
        self._register_adapter(ElementAdapter, self.elements, *args, **kwargs)

    def register_command(self, *args, **kwargs):
        """Add parser and generator support for a command.

        Other Parameters
        ----------------
        aliases : str or sequence
            The command alias(es).

        kwargs : mapping, optional
            Keyword arguments to pass to the adapter constructor; defaults to None.

        overwrite : bool, optional
            Overwrite commands with the same aliases, if present. Defaults to False.
        """
        self._register_adapter(CommandAdapter, self.commands, *args, **kwargs)

    def register_analysis(self, *args, **kwargs):
        """Add parser and generator support for an analysis.

        Other Parameters
        ----------------
        aliases : str or sequence
            The analysis alias(es).

        kwargs : mapping, optional
            Keyword arguments to pass to the adapter constructor; defaults to None.

        overwrite : bool, optional
            Overwrite analyses with the same aliases, if present. Defaults to False.
        """
        self._register_adapter(AnalysisAdapter, self.analyses, *args, **kwargs)

    def adapter_by_setter(self, setter):
        """Get adapter given its Python setter.

        Parameters
        ----------
        setter : type
            The setter to look up the adapter for.

        Returns
        -------
        :class:`.BaseAdapter`
            The adapter corresponding to `setter`.

        Raises
        ------
        ValueError
            If no adapter corresponding to `setter` could be found.
        """
        for adapter in self.directives.values():
            if adapter.setter == setter:
                return adapter

        raise ValueError(f"Could not find adapter for '{setter!r}'")

    def match_fuzzy_directive(self, search, limit=3, cutoff=0.5):
        """Get the directives that most closely match the specified string.

        Parameters
        ----------
        search : str
            The directive to search for.

        limit : int, optional
            The maximum number of matches to return.

        cutoff : float, optional
            The cutoff below which to assume no match. This is the ratio as defined in
            the `Python documentation
            <https://docs.python.org/3/library/difflib.html#difflib.SequenceMatcher.ratio>`__.

        Returns
        -------
        list
            Up to `limit` closest matches.
        """
        return get_close_matches(search, self.directives, n=limit, cutoff=cutoff)


class KatSpec(BaseSpec):
    """Kat language specification.

    This defines the available instructions for the parser and the adapters that convert
    them to and from Python models. The default instructions, actions and keywords are
    built into public properties which may be modified by users (e.g. to add support for
    custom commands). As such, the internal defaults (fields with names beginning
    `_DEFAULT_`) should not be modified after import.
    """

    # List of default elements in (aliases, type, kwargs) form.
    # :class:`.InstructionAdapter` objects are created for each element and the aliases
    # are each mapped to their corresponding adapter. Order here does not matter.
    _DEFAULT_ELEMENTS = [
        # Components.
        (
            ("amplifier", "amp"),
            {"setter": electronics.Amplifier, "getter": electronics.Amplifier},
        ),
        (
            ("beamsplitter", "bs"),
            {"setter": components.Beamsplitter, "getter": components.Beamsplitter},
        ),
        # Cavity's `build_last` flag is set because it implicitly depends on any nodes
        # that appear in the path from its start port back to itself, so its
        # dependencies cannot be determined by the time the first set of elements are
        # built into the model. It is therefore moved to the second build pass by this
        # flag.
        (
            ("cavity", "cav"),
            {
                "setter": components.Cavity,
                "getter": components.Cavity,
                "build_last": True,
            },
        ),
        (
            ("degree_of_freedom", "dof"),
            {
                "setter": components.DegreeOfFreedom,
                "getter": components.DegreeOfFreedom,
            },
        ),
        (
            ("directional_beamsplitter", "dbs"),
            {
                "setter": components.DirectionalBeamsplitter,
                "getter": components.DirectionalBeamsplitter,
            },
        ),
        (
            ("filter_zpk", "zpk"),
            {"setter": electronics.ZPKFilter, "getter": electronics.ZPKFilter},
        ),
        (
            ("filter_butter", "butter"),
            {"setter": electronics.ButterFilter, "getter": electronics.ButterFilter},
        ),
        (
            ("filter_cheby1", "cheby1"),
            {"setter": electronics.Cheby1Filter, "getter": electronics.Cheby1Filter},
        ),
        (
            ("isolator", "isol"),
            {"setter": components.Isolator, "getter": components.Isolator},
        ),
        (("laser", "l"), {"setter": components.Laser, "getter": components.Laser}),
        ("lens", {"setter": components.Lens, "getter": components.Lens}),
        (("mirror", "m"), {"setter": components.Mirror, "getter": components.Mirror}),
        (
            ("modulator", "mod"),
            {"setter": components.Modulator, "getter": components.Modulator},
        ),
        (
            ("optical_bandpass", "obp"),
            {
                "setter": components.optical_bandpass.OpticalBandpassFilter,
                "getter": components.optical_bandpass.OpticalBandpassFilter,
            },
        ),
        (
            ("squeezer", "sq"),
            {"setter": components.Squeezer, "getter": components.Squeezer},
        ),
        (
            "readout_dc",
            {"setter": components.ReadoutDC, "getter": components.ReadoutDC},
        ),
        (
            "readout_dc_qpd",
            {"setter": components.ReadoutDCQPD, "getter": components.ReadoutDCQPD},
        ),
        (
            "readout_rf",
            {"setter": components.ReadoutRF, "getter": components.ReadoutRF},
        ),
        (
            ("variable", "var"),
            {"setter": components.Variable, "getter": components.Variable},
        ),
        # Detectors.
        (
            ("amplitude_detector", "ad"),
            {
                "setter": detectors.AmplitudeDetector,
                "getter": detectors.AmplitudeDetector,
            },
        ),
        (
            "astigd",
            {
                "setter": detectors.AstigmatismDetector,
                "getter": detectors.AstigmatismDetector,
            },
        ),
        (
            ("beam_property_detector", "bp"),
            {
                "setter": detectors.BeamPropertyDetector,
                "getter": detectors.BeamPropertyDetector,
            },
        ),
        ("ccd", {"setter": detectors.CCD, "getter": detectors.CCD}),
        ("ccdline", {"setter": detectors.CCDScanLine, "getter": detectors.CCDScanLine}),
        ("ccdpx", {"setter": detectors.CCDPixel, "getter": detectors.CCDPixel}),
        (
            "cp",
            {
                "setter": detectors.CavityPropertyDetector,
                "getter": detectors.CavityPropertyDetector,
            },
        ),
        ("fcam", {"setter": detectors.FieldCamera, "getter": detectors.FieldCamera}),
        (
            "fline",
            {"setter": detectors.FieldScanLine, "getter": detectors.FieldScanLine},
        ),
        ("fpx", {"setter": detectors.FieldPixel, "getter": detectors.FieldPixel}),
        # Gouy's `build_last` flag is set because it implicitly depends on any nodes
        # that appear in the path from its start port back to itself, so its
        # dependencies cannot be determined by the time the first set of elements are
        # built into the model. It is therefore moved to the second build pass by this
        # flag.
        (
            "gouy",
            {"setter": detectors.Gouy, "getter": detectors.Gouy, "build_last": True},
        ),
        ("knmd", {"setter": detectors.KnmDetector, "getter": detectors.KnmDetector}),
        (
            "mmd",
            {
                "setter": detectors.ModeMismatchDetector,
                "getter": detectors.ModeMismatchDetector,
            },
        ),
        (
            ("motion_detector", "xd"),
            {"setter": detectors.MotionDetector, "getter": detectors.MotionDetector},
        ),
        (
            ("power_detector_dc", "pd"),
            {"setter": detectors.PowerDetector, "getter": detectors.PowerDetector},
        ),
        (
            ("power_detector_demod_1", "pd1"),
            {
                "setter": detectors.PowerDetectorDemod1,
                "getter": detectors.PowerDetectorDemod1,
            },
        ),
        (
            ("power_detector_demod_2", "pd2"),
            {
                "setter": detectors.PowerDetectorDemod2,
                "getter": detectors.PowerDetectorDemod2,
            },
        ),
        (
            ("quantum_noise_detector", "qnoised"),
            {
                "setter": detectors.QuantumNoiseDetector,
                "getter": detectors.QuantumNoiseDetector,
            },
        ),
        (
            ("quantum_noise_detector_demod_1", "qnoised1"),
            {
                "setter": detectors.QuantumNoiseDetectorDemod1,
                "getter": detectors.QuantumNoiseDetectorDemod1,
            },
        ),
        (
            ("quantum_noise_detector_demod_2", "qnoised2"),
            {
                "setter": detectors.QuantumNoiseDetectorDemod2,
                "getter": detectors.QuantumNoiseDetectorDemod2,
            },
        ),
        (
            ("quantum_shot_noise_detector", "qshot"),
            {
                "setter": detectors.QuantumShotNoiseDetector,
                "getter": detectors.QuantumShotNoiseDetector,
            },
        ),
        (
            ("quantum_shot_noise_detector_demod_1", "qshot1"),
            {
                "setter": detectors.QuantumShotNoiseDetectorDemod1,
                "getter": detectors.QuantumShotNoiseDetectorDemod1,
            },
        ),
        (
            ("quantum_shot_noise_detector_demod_2", "qshot2"),
            {
                "setter": detectors.QuantumShotNoiseDetectorDemod2,
                "getter": detectors.QuantumShotNoiseDetectorDemod2,
            },
        ),
        (
            ("signal_generator", "sgen"),
            {
                "setter": components.SignalGenerator,
                "getter": components.SignalGenerator,
            },
        ),
        ("splitpd", {"setter": detectors.SplitPD, "getter": detectors.SplitPD}),
        (
            ("zpk_actuator", "actuator"),
            {
                "setter": electronics.ZPKNodeActuator,
                "getter": electronics.ZPKNodeActuator,
            },
        ),
        # Connectors.
        (("space", "s"), {"setter": components.Space, "getter": components.Space}),
        ("nothing", {"setter": components.Nothing, "getter": components.Nothing}),
        # Mechanics.
        ("free_mass", {"setter": mechanical.FreeMass, "getter": mechanical.FreeMass}),
        ("pendulum", {"setter": mechanical.Pendulum, "getter": mechanical.Pendulum}),
        (
            "ligo_triple",
            {"setter": ligo.LIGOTripleSuspension, "getter": ligo.LIGOTripleSuspension},
        ),
        (
            "ligo_quad",
            {"setter": ligo.LIGOQuadSuspension, "getter": ligo.LIGOQuadSuspension},
        ),
        # Lock.
        ("lock", {"setter": locks.Lock, "getter": locks.Lock}),
        # Noises.
        ("noise", {"setter": noise.ClassicalNoise, "getter": noise.ClassicalNoise}),
        # Gauss.
        ("gauss", {"setter": components.Gauss, "getter": _GuassGetterProxy()},),
    ]

    # List of default function adapters.
    _DEFAULT_COMMANDS = [
        (
            # Fsig.
            # This technically sets a component (:class:`.finesse.frequency.Fsig`), but
            # it's always present in models so this is instead implemented as a command.
            "fsig",
            {"setter": _set_fsig, "getter": _FsigGetterProxy(), "singular": True},
        ),
        (
            "lambda",
            {"setter": _set_lambda0, "getter": _LambdaGetterProxy(), "singular": True},
        ),
        (
            "modes",
            {
                "setter": Model.select_modes,
                "getter": _ModesGetterProxy(),
                "singular": True,
            },
        ),
        ("link", {"setter": _set_link, "singular": False},),
        (
            "intrix",
            {"setter": _set_intrix, "getter": _IntrixGetterProxy(), "singular": False},
        ),
        ("tem", {"setter": _set_tem, "getter": _TEMGetterProxy(), "singular": False},),
    ]

    _DEFAULT_ANALYSES = [
        # Group actions.
        ("parallel", {"setter": actions.Parallel, "getter": actions.Parallel}),
        ("series", {"setter": actions.Series, "getter": actions.Series}),
        # Axes.
        ("noxaxis", {"setter": actions.Noxaxis, "getter": actions.Noxaxis}),
        ("xaxis", {"setter": actions.Xaxis, "getter": actions.Xaxis}),
        ("x2axis", {"setter": actions.X2axis, "getter": actions.X2axis}),
        ("x3axis", {"setter": actions.X3axis, "getter": actions.X3axis}),
        ("sweep", {"setter": actions.Sweep, "getter": actions.Sweep}),
        ("change", {"setter": actions.Change, "getter": actions.Change}),
        (
            ("freqresp", "frequency_response"),
            {"setter": actions.FrequencyResponse, "getter": actions.FrequencyResponse},
        ),
        (
            "opt_rf_readout_phase",
            {
                "setter": actions.OptimiseRFReadoutPhaseDC,
                "getter": actions.OptimiseRFReadoutPhaseDC,
            },
        ),
        (
            "sensing_matrix_dc",
            {"setter": actions.SensingMatrixDC, "getter": actions.SensingMatrixDC},
        ),
        # Model physics.
        (
            "noise_analysis",
            {"setter": noise.NoiseAnalysis, "getter": noise.NoiseAnalysis},
        ),
        ("abcd", {"setter": actions.ABCD, "getter": actions.ABCD}),
        ("beam_trace", {"setter": actions.BeamTrace, "getter": actions.BeamTrace}),
        (
            "propagate_beam",
            {"setter": actions.PropagateBeam, "getter": actions.PropagateBeam},
        ),
        (
            "propagate_beam_astig",
            {
                "setter": actions.PropagateAstigmaticBeam,
                "getter": actions.PropagateAstigmaticBeam,
            },
        ),
        # Utilities.
        ("debug", {"setter": actions.Debug, "getter": actions.Debug}),
        ("plot", {"setter": actions.Plot, "getter": actions.Plot}),
        ("print", {"setter": actions.Printer, "getter": actions.Printer}),
        ("run_locks", {"setter": actions.RunLocks, "getter": actions.RunLocks}),
        (
            "noise_projection",
            {"setter": actions.NoiseProjection, "getter": actions.NoiseProjection},
        ),
        ("print_model", {"setter": actions.PrintModel, "getter": actions.PrintModel}),
        (
            "print_model_attr",
            {"setter": actions.PrintModelAttr, "getter": actions.PrintModelAttr},
        ),
    ]

    _SUPPORTED_KEYWORDS = {
        # None.
        "none",
        # HOM collections.
        "even",
        "odd",
        "x",
        "y",
        "off",
        # Axis scales.
        "lin",
        "log",
        # Modulator types.
        "am",
        "pm",
        # Filter types.
        "lowpass",
        "highpass",
        "bandpass",
        "bandstop",
        "xsplit",
        "ysplit",
        # Beam properties (see :class:`finesse.detectors.compute.gaussian.BeamProperty`).
        *detectors.bpdetector.BP_KEYWORDS.keys(),
        # Cavity properties (see :class:`finesse.detectors.compute.gaussian.CavityProperty`).
        *detectors.cavity_detector.CP_KEYWORDS.keys(),
    }

    _SUPPORTED_CONSTANTS = symbols.CONSTANTS

    _SUPPORTED_UNARY_OPERATORS = {
        "+": symbols.FUNCTIONS["pos"],
        "-": symbols.FUNCTIONS["neg"],
    }

    _SUPPORTED_BINARY_OPERATORS = {
        "+": symbols.OPERATORS["__add__"],
        "-": symbols.OPERATORS["__sub__"],
        "*": symbols.OPERATORS["__mul__"],
        "**": symbols.OPERATORS["__pow__"],
        "/": symbols.OPERATORS["__truediv__"],
        "//": symbols.OPERATORS["__floordiv__"],
    }

    # Built-in functions.
    _SUPPORTED_EXPRESSION_FUNCTIONS = symbols.FUNCTIONS



import pytest


@pytest.fixture
def tem_model(model):
    model.parse(
        """
        l L0 P=1
        gauss gL0 L0.p1.o w0=1m z=0
        modes(x, maxtem=1)
        """
    )
    return model


def test_tem_default(tem_model):
    assert not tem_model.L0.non_default_power_coeffs


def test_tem_default_mode_nondefault_factor(tem_model):
    """Default modes with non-default factor."""
    tem_model.parse("tem(L0, 0, 0, factor=0.5)")
    assert tem_model.L0.power_coeffs == {(0, 0): (0.5, 0)}


def test_tem(tem_model):
    tem_model.parse(
        """
        tem(L0, 0, 0, factor=0)
        tem(L0, 1, 0, factor=1)
        tem(L0, 1, 1, factor=0.5, phase=35)
        """
    )

    assert tem_model.L0.power_coeffs == {
        (0, 0): (0, 0),
        (1, 0): (1, 0),
        (1, 1): (0.5, 35),
    }



"""
Plane wave radiation pressure tests. For derivations of the analytics used here, see
https://finesse.readthedocs.io/en/latest/developer/testing/analytics/radiation_pressure.html
"""

import pytest
import numpy as np


@pytest.fixture
def mirror_radiation_pressure_single_side_model(model):
    model.parse(
        """
        l l1 P=3
        s s1 l1.p1 m1.p1 L=5
        m m1 R=0.5 T=0.4
        free_mass m1_sus m1.mech 0.5e-3

        ad upper s1.p1.o &fsig.f
        ad lower s1.p1.o -&fsig.f

        sgen sig l1.amp.i 0.1

        fsig(0.1)
        xaxis(sig.f, log, 0.1, 1e7, 400)
    """
    )

    return model


@pytest.fixture
def mirror_radiation_pressure_two_sides_one_modulation_model(model):
    model.parse(
        """
        l l1 P=3
        s s1 l1.p1 m1.p1 L=5
        m m1 R=1 T=0
        s s2 m1.p2 l2.p1 L=5
        l l2 P=2

        free_mass m1_sus m1.mech 0.5e-3

        ad upper s2.p2.o &fsig.f
        ad lower s2.p2.o -&fsig.f

        sgen sig l1.amp.i 0.1

        fsig(0.1)
        xaxis(sig.f, log, 0.1, 1e7, 400)
    """
    )

    return model


@pytest.fixture
def mirror_radiation_pressure_two_sides_two_modulations_model(model):
    model.parse(
        """
        l l1 P=3
        s s1 l1.p1 m1.p1 L=5
        m m1 R=1 T=0
        s s2 m1.p2 l2.p1 L=5
        l l2 P=4

        free_mass m1_sus m1.mech 0.5e-3

        ad upper_left s1.p1.o &fsig.f
        ad lower_left s1.p1.o -&fsig.f

        ad upper_right s2.p2.o &fsig.f
        ad lower_right s2.p2.o -&fsig.f

        sgen sig1 l1.amp.i 0.1
        sgen sig2 l2.amp.i 0.1 phase=29

        fsig(0.1)
        xaxis(fsig.f, log, 0.1, 1e7, 400)
    """
    )

    return model


def test_mirror_radiation_pressure_single_side(
    mirror_radiation_pressure_single_side_model,
):
    """
    Test modulation of signal sidebands by radiation pressure forces on a mirror's front side.
    """
    import sympy as sy
    import scipy.constants as constants

    model = mirror_radiation_pressure_single_side_model
    out = model.run()

    c, L, P, m, n, R, T, omega, lambda0 = sy.var(
        "c L P m n R T omega lambda0", real=True
    )
    tau = L / c
    omega_0 = 2 * sy.pi * c / lambda0
    r = sy.sqrt(R)

    # Effective modulation index due to radiation pressure feedback
    rp_mod_index = 2 * n * (1 + R - T) * P * omega_0 / (m * c ** 2 * omega ** 2)

    upper = r * sy.sqrt(P) * sy.exp(-2j * omega * tau) * (n + 2j * rp_mod_index) / 4
    lower = r * sy.sqrt(P) * sy.exp(2j * omega * tau) * (n + 2j * rp_mod_index) / 4

    values = {
        c: constants.c,
        L: model.elements["s1"].L,
        P: model.l1.P,
        m: model.m1_sus.mass,
        R: model.m1.R,
        T: model.m1.T,
        n: model.sig.amplitude,
        lambda0: model.lambda0,
    }

    upper = sy.lambdify(omega, upper.subs(values))(2 * np.pi * out.x[0])
    lower = sy.lambdify(omega, lower.subs(values))(2 * np.pi * out.x[0])

    assert np.all(abs(upper - out["upper"]) / abs(upper) < 1e-13)
    assert np.all(abs(lower - out["lower"]) / abs(lower) < 1e-13)


def test_mirror_radiation_pressure_two_sides_one_mod(
    mirror_radiation_pressure_two_sides_one_modulation_model,
):
    """
    Test creation of signal sidebands on a mirror's back side, by
    radiation pressure forces on a mirror's front side.
    """
    import sympy as sy
    import scipy.constants as constants

    model = mirror_radiation_pressure_two_sides_one_modulation_model
    out = model.run()

    c, L2, P1, P2, m, n, R, omega, lambda0 = sy.var(
        "c L2 P1 P2 m n R omega lambda0", real=True
    )
    tau = L2 / c
    omega_0 = 2 * sy.pi * c / lambda0
    r = sy.sqrt(R)

    # Effective modulation index due to radiation pressure feedback
    rp_mod_index = 2 * n * (1 + R) * P1 * omega_0 / (m * c ** 2 * omega ** 2)

    upper = r * sy.sqrt(P2) * sy.exp(-2j * omega * tau) * (-1j * rp_mod_index) / 2
    lower = r * sy.sqrt(P2) * sy.exp(2j * omega * tau) * (-1j * rp_mod_index) / 2

    values = {
        c: constants.c,
        L2: model.elements["s2"].L,
        P1: model.l1.P,
        P2: model.l2.P,
        m: model.m1_sus.mass,
        R: model.m1.R,
        n: model.sig.amplitude,
        lambda0: model.lambda0,
    }

    upper = sy.lambdify(omega, upper.subs(values))(2 * np.pi * out.x[0])
    lower = sy.lambdify(omega, lower.subs(values))(2 * np.pi * out.x[0])

    assert np.all(abs(upper - out["upper"]) / abs(upper) < 1e-13)
    assert np.all(abs(lower - out["lower"]) / abs(lower) < 1e-13)


def test_mirror_radiation_pressure_two_sides_two_mod(
    mirror_radiation_pressure_two_sides_two_modulations_model,
):
    """
    Test modulation of signal sidebands by radiation pressure forces on both sides of a mirror.
    """
    import sympy as sy
    import scipy.constants as constants

    # N.B. This test assumes model.s1.L == model.s2.L
    model = mirror_radiation_pressure_two_sides_two_modulations_model
    out = model.run()

    c, L, P1, P2, m, n1, n2, psi, R, omega, lambda0 = sy.var(
        "c L P1 P2 m n1 n2 psi R omega lambda0", real=True
    )
    tau = L / c
    omega_0 = 2 * sy.pi * c / lambda0
    r = sy.sqrt(R)

    D = sy.sqrt(
        n1 ** 2 * P1 ** 2 - 2 * n1 * n2 * P1 * P2 * sy.cos(psi) + n2 ** 2 * P2 ** 2
    )
    theta_left = sy.atan2(n2 * P2 * sy.sin(psi), n1 * P1 - n2 * P2 * sy.cos(psi))
    theta_right = sy.atan2(n1 * P1 * sy.sin(-psi), n2 * P2 - n1 * P1 * sy.cos(-psi))
    # Effective modulation index due to radiation pressure feedback
    rp_mod_index = 1j * (1 + R) * D * omega_0 / (m * c ** 2 * omega ** 2)

    upper_left = (
        r
        * sy.sqrt(P1)
        * (
            (n1 / 4) * sy.exp(-1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(-1j * (2 * omega * tau + theta_left))
        )
    )
    lower_left = (
        r
        * sy.sqrt(P1)
        * (
            (n1 / 4) * sy.exp(1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(1j * (2 * omega * tau + theta_left))
        )
    )

    upper_right = (
        r
        * sy.sqrt(P2)
        * (
            (n1 / 4) * sy.exp(-1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(-1j * (2 * omega * tau + theta_right))
        )
        * sy.exp(1j * psi)
    )
    lower_right = (
        r
        * sy.sqrt(P2)
        * (
            (n1 / 4) * sy.exp(1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(1j * (2 * omega * tau + theta_right))
        )
        * sy.exp(-1j * psi)
    )

    values = {
        c: constants.c,
        L: model.elements["s1"].L,
        P1: model.l1.P,
        P2: model.l2.P,
        m: model.m1_sus.mass,
        R: model.m1.R,
        n1: model.sig1.amplitude,
        n2: model.sig2.amplitude,
        psi: np.radians(model.sig2.phase.value),
        lambda0: model.lambda0,
    }

    upper_left = sy.lambdify(omega, upper_left.subs(values))(2 * np.pi * out.x[0])
    lower_left = sy.lambdify(omega, lower_left.subs(values))(2 * np.pi * out.x[0])
    upper_right = sy.lambdify(omega, upper_right.subs(values))(2 * np.pi * out.x[0])
    lower_right = sy.lambdify(omega, lower_right.subs(values))(2 * np.pi * out.x[0])

    assert np.all(abs(upper_left - out["upper_left"]) / abs(upper_left) < 1e-13)
    assert np.all(abs(lower_left - out["lower_left"]) / abs(lower_left) < 1e-13)
    assert np.all(abs(upper_right - out["upper_right"]) / abs(upper_right) < 1e-13)
    assert np.all(abs(lower_right - out["lower_right"]) / abs(lower_right) < 1e-13)


# FIXME: The following beamsplitter tests will all break until
# beamsplitter signal calculations are done


@pytest.fixture
def beamsplitter_radiation_pressure_single_side_model(model):
    model.parse(
        """
        l l1 P=3
        s s1 l1.p1 bs1.p1 L=5
        bs bs1 R=0.5 T=0.4 alpha=37
        s s2 bs1.p2 out.p1 L=5
        nothing out

        free_mass bs1_sus bs1.mech 0.5e-3

        ad upper s2.p2.o &fsig.f
        ad lower s2.p2.o -&fsig.f

        sgen sig l1.amp.i 0.1

        fsig(0.1)
        xaxis(fsig.f, log, 0.1, 1e7, 400)
        """
    )

    return model


@pytest.fixture
def beamsplitter_radiation_pressure_two_sides_one_modulation_model(model):
    model.parse(
        """
        l l1 P=3
        s s1 l1.p1 bs1.p1 L=5

        l l2 P=2
        s s2 l2.p1 bs1.p4 L=5

        bs bs1 R=1 T=0 alpha=37
        s sout bs1.p3 out.p1 L=5
        nothing out

        free_mass bs1_sus bs1.mech 0.5e-3

        ad upper sout.p2.o &fsig.f
        ad lower sout.p2.o -&fsig.f

        sgen sig l1.amp.i 0.1

        fsig(0.1)
        xaxis(fsig.f, log, 0.1, 1e7, 400)
    """
    )

    return model


@pytest.fixture
def beamsplitter_radiation_pressure_two_sides_two_modulations_model(model):
    model.parse(
        """
        l l1 P=3
        s s1 l1.p1 bs1.p1 L=5

        l l2 P=4
        s s2 l2.p1 bs1.p4 L=5

        bs bs1 R=1 T=0 alpha=37

        s sout_left bs1.p2 out_left.p1 L=5
        nothing out_left

        s sout_right bs1.p3 out_right.p1 L=5
        nothing out_right

        free_mass bs1_sus bs1.mech 0.5e-3

        ad upper_left sout_left.p2.o &fsig.f
        ad lower_left sout_left.p2.o -&fsig.f

        ad upper_right sout_right.p2.o &fsig.f
        ad lower_right sout_right.p2.o -&fsig.f

        sgen sig1 l1.amp.i 0.1
        sgen sig2 l2.amp.i 0.1 phase=29

        fsig(0.1)
        xaxis(fsig.f, log, 0.1, 1e7, 400)
    """
    )

    return model


def test_beamsplitter_radiation_pressure_single_side(
    beamsplitter_radiation_pressure_single_side_model,
):
    """
    Test modulation of signal sidebands by radiation pressure forces on a beamsplitter's front
    side.
    """
    import sympy as sy
    import scipy.constants as constants

    model = beamsplitter_radiation_pressure_single_side_model
    out = model.run()

    c, L, P, m, n, R, T, omega, lambda0, alpha = sy.var(
        "c L P m n R T omega lambda0 alpha", real=True
    )
    tau = L / c
    omega_0 = 2 * sy.pi * c / lambda0
    r = sy.sqrt(R)

    # Effective modulation index due to radiation pressure feedback
    rp_mod_index = (
        2
        * n
        * (1 + R - T)
        * P
        * omega_0
        / (m * c ** 2 * omega ** 2)
        * sy.cos(alpha) ** 2
    )

    upper = r * sy.sqrt(P) * sy.exp(-2j * omega * tau) * (n + 2j * rp_mod_index) / 4
    lower = r * sy.sqrt(P) * sy.exp(2j * omega * tau) * (n + 2j * rp_mod_index) / 4

    values = {
        c: constants.c,
        L: model.elements["s1"].L,
        P: model.l1.P,
        m: model.bs1_sus.mass,
        R: model.bs1.R,
        T: model.bs1.T,
        alpha: np.radians(model.bs1.alpha.value),
        n: model.sig.amplitude,
        lambda0: model.lambda0,
    }

    upper = sy.lambdify(omega, upper.subs(values))(2 * np.pi * out.x[0])
    lower = sy.lambdify(omega, lower.subs(values))(2 * np.pi * out.x[0])

    assert np.all(abs(upper - out["upper"]) / abs(upper) < 1e-13)
    assert np.all(abs(lower - out["lower"]) / abs(lower) < 1e-13)


def test_beamsplitter_radiation_pressure_two_sides_one_mod(
    beamsplitter_radiation_pressure_two_sides_one_modulation_model,
):
    """
    Test creation of signal sidebands on a beamsplitter's back side, by radiation pressure forces
    on the front side.
    """
    import sympy as sy
    import scipy.constants as constants

    model = beamsplitter_radiation_pressure_two_sides_one_modulation_model
    out = model.run()

    c, L2, P1, P2, m, n, R, omega, lambda0, alpha = sy.var(
        "c L2 P1 P2 m n R omega lambda0 alpha", real=True
    )
    tau = L2 / c
    omega_0 = 2 * sy.pi * c / lambda0
    r = sy.sqrt(R)

    # Effective modulation index due to radiation pressure feedback
    rp_mod_index = (
        2 * n * (1 + R) * P1 * omega_0 / (m * c ** 2 * omega ** 2) * sy.cos(alpha) ** 2
    )

    upper = r * sy.sqrt(P2) * sy.exp(-2j * omega * tau) * (-1j * rp_mod_index) / 2
    lower = r * sy.sqrt(P2) * sy.exp(2j * omega * tau) * (-1j * rp_mod_index) / 2

    values = {
        c: constants.c,
        L2: model.elements["s2"].L,
        P1: model.l1.P,
        P2: model.l2.P,
        m: model.bs1_sus.mass,
        R: model.bs1.R,
        alpha: np.radians(model.bs1.alpha.value),
        n: model.sig.amplitude,
        lambda0: model.lambda0,
    }

    upper = sy.lambdify(omega, upper.subs(values))(2 * np.pi * out.x[0])
    lower = sy.lambdify(omega, lower.subs(values))(2 * np.pi * out.x[0])

    assert np.all(abs(upper - out["upper"]) / abs(upper) < 1e-13)
    assert np.all(abs(lower - out["lower"]) / abs(lower) < 1e-13)


def test_beamsplitter_radiation_pressure_two_sides_two_mod(
    beamsplitter_radiation_pressure_two_sides_two_modulations_model,
):
    """
    Test modulation of signal sidebands by radiation pressure forces on both sides of a
    beamsplitter.
    """
    import sympy as sy
    import scipy.constants as constants

    # N.B. This test assumes model.s1.L == model.s2.L
    model = beamsplitter_radiation_pressure_two_sides_two_modulations_model
    out = model.run()

    c, L, P1, P2, m, n1, n2, psi, R, omega, lambda0, alpha = sy.var(
        "c L P1 P2 m n1 n2 psi R omega lambda0 alpha", real=True
    )
    tau = L / c
    omega_0 = 2 * sy.pi * c / lambda0
    r = sy.sqrt(R)

    D = sy.sqrt(
        n1 ** 2 * P1 ** 2 - 2 * n1 * n2 * P1 * P2 * sy.cos(psi) + n2 ** 2 * P2 ** 2
    )
    theta_left = sy.atan2(n2 * P2 * sy.sin(psi), n1 * P1 - n2 * P2 * sy.cos(psi))
    theta_right = sy.atan2(n1 * P1 * sy.sin(-psi), n2 * P2 - n1 * P1 * sy.cos(-psi))
    # Effective modulation index due to radiation pressure feedback
    rp_mod_index = (
        1j * (1 + R) * D * omega_0 / (m * c ** 2 * omega ** 2) * sy.cos(alpha) ** 2
    )

    upper_left = (
        r
        * sy.sqrt(P1)
        * (
            (n1 / 4) * sy.exp(-1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(-1j * (2 * omega * tau + theta_left))
        )
    )
    lower_left = (
        r
        * sy.sqrt(P1)
        * (
            (n1 / 4) * sy.exp(1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(1j * (2 * omega * tau + theta_left))
        )
    )

    upper_right = (
        r
        * sy.sqrt(P2)
        * (
            (n1 / 4) * sy.exp(-1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(-1j * (2 * omega * tau + theta_right))
        )
        * sy.exp(1j * psi)
    )
    lower_right = (
        r
        * sy.sqrt(P2)
        * (
            (n1 / 4) * sy.exp(1j * (2 * omega * tau))
            + rp_mod_index * sy.exp(1j * (2 * omega * tau + theta_right))
        )
        * sy.exp(-1j * psi)
    )

    values = {
        c: constants.c,
        L: model.elements["s1"].L,
        P1: model.l1.P,
        P2: model.l2.P,
        m: model.bs1_sus.mass,
        R: model.bs1.R,
        alpha: np.radians(model.bs1.alpha.value),
        n1: model.sig1.amplitude,
        n2: model.sig2.amplitude,
        psi: np.radians(model.sig2.phase.value),
        lambda0: model.lambda0,
    }

    upper_left = sy.lambdify(omega, upper_left.subs(values))(2 * np.pi * out.x[0])
    lower_left = sy.lambdify(omega, lower_left.subs(values))(2 * np.pi * out.x[0])
    upper_right = sy.lambdify(omega, upper_right.subs(values))(2 * np.pi * out.x[0])
    lower_right = sy.lambdify(omega, lower_right.subs(values))(2 * np.pi * out.x[0])

    assert np.all(abs(upper_left - out["upper_left"]) / abs(upper_left) < 1e-13)
    assert np.all(abs(lower_left - out["lower_left"]) / abs(lower_left) < 1e-13)
    assert np.all(abs(upper_right - out["upper_right"]) / abs(upper_right) < 1e-13)
    assert np.all(abs(lower_right - out["lower_right"]) / abs(lower_right) < 1e-13)



"""Equality functions for Finesse objects.

There is a lot going on in this module. We cannot compare some Finesse objects by implementing
`__eq__` methods directly, so special checks are implemented here. Furthermore, in order to provide
useful error messages for tests, special set operations are provided that use the special
equivalence functions here instead of hashing.
"""

from io import BytesIO
from numpy.testing import assert_array_equal
import networkx as nx


def assert_graph_matches_def(graph, ref_definition):
    """Assert specified graph has GML representation equivalent to specified definition."""
    ref_graph = nx.read_gml(BytesIO(ref_definition.encode()))
    assert_graphs_match(graph, ref_graph)


def assert_graphs_match(graph, ref_graph):
    """Assert graphs match."""

    def stringify(item):
        """Convert graph item to string form."""
        if isinstance(item, list):
            return f"[{', '.join(stringify(i) for i in item)}]"
        return str(item)

    nodes = sorted([node for node in graph.nodes], key=lambda node: str(node))
    ref_nodes = sorted([node for node in ref_graph.nodes], key=lambda node: str(node))

    # Node names.
    assert [stringify(node) for node in nodes] == [
        stringify(node) for node in ref_nodes
    ], "Node names in graph don't match those in reference graph"

    # Node attributes.
    for node, ref_node in zip(nodes, ref_nodes):
        data = {
            stringify(key): stringify(value) for key, value in graph.nodes[node].items()
        }
        ref_data = {
            stringify(key): stringify(value)
            for key, value in ref_graph.nodes[ref_node].items()
        }
        assert (
            data == ref_data
        ), f"{node!r} attributes in graph don't match those in reference graph"

    edges = sorted(
        [edge for edge in graph.edges],
        key=lambda edge: (stringify(edge[0]), stringify(edge[1])),
    )
    ref_edges = sorted(
        [edge for edge in ref_graph.edges],
        key=lambda edge: (stringify(edge[0]), stringify(edge[1])),
    )

    # Edges.
    assert list(edges) == list(ref_edges)

    # Edge attributes.
    for edge, ref_edge in zip(edges, ref_edges):
        data = {
            stringify(key): stringify(value) for key, value in graph.edges[edge].items()
        }
        ref_data = {
            stringify(key): stringify(value)
            for key, value in ref_graph.edges[ref_edge].items()
        }
        assert (
            data == ref_data
        ), f"{edge!r} attributes in graph don't match those in reference graph"


def assert_models_equivalent(model_a, model_b):
    """Check :class:`.Model` equality.

    This checks that the two specified models are identical in a shallow way, i.e. they have
    elements with the same properties, equivalent networks, actions, etc.

    This is implemented here and not as `Model.__eq__` because :class:`.Model` must be hashable.
    """
    if model_a.analysis is not None and model_b.analysis is not None:
        assert_actions_equivalent(model_a.analysis, model_b.analysis)
    elif (model_a.analysis is None) != (model_b.analysis is None):
        raise AssertionError("analyses don't match")

    assert set(model_a.elements) == set(model_b.elements)
    assert_array_equal(model_a.homs, model_b.homs, "HOMs don't match")

    for el_a, el_b in zip(sorted(model_a.elements), sorted(model_b.elements)):
        assert_model_elements_equivalent(model_a.elements[el_a], model_b.elements[el_b])


def assert_model_elements_equivalent(element_a, element_b):
    """Check :class:`.ModelElement` equality.

    This is implemented here and not as `ModelElement.__eq__` because `ModelElement` must be
    hashable.
    """

    assert type(element_a) == type(element_b)
    assert element_a.name == element_a.name
    for param_a, param_b in zip(element_a.parameters, element_b.parameters):
        assert_parameters_equivalent(param_a, param_b)


def assert_parameters_equivalent(param_a, param_b, expected=None):
    """Check model parameter equality.

    This is implemented here and not as `ModelParameter.__eq__` because `ModelParameter` must be
    hashable.
    """
    assert type(param_a) == type(param_b)

    val_a = param_a.eval()
    val_b = param_b.eval()

    assert val_a == val_b

    if expected is not None:
        assert val_a == expected
        assert val_b == expected


# FIXME: make this check far more than just the name of the top action...
def assert_actions_equivalent(act_a, act_b):
    """Check action equality.

    This is implemented here and not as `Action.__eq__` because `Action` must be hashable.
    """
    # TODO: compare more things.
    assert str(act_a.plan()) == str(act_b.plan())



# import logging
# import numpy as np
# import re

# from finesse.components.node import NodeType
# from finesse.simulations.base import BaseSimulation

# LOGGER = logging.getLogger(__name__)


# class DebugSimulation(BaseSimulation):
#     def active(self):
#         return self._M is not None

#     def print_matrix(self):
#         self._M.print_matrix()

#     def solve(self):
#         self.out = None
#         return

#     def __enter__(self):
#         """
#         When entered the Simulation object will create the matrix to be used in
#         the simulation.

#         This is where the matrix gets allocated and constructed. It will expect
#         that the model structure does not change after this until it has been
#         exited.
#         """
#         # Initialising the simulation expects there to be a self._M class that handles the
#         # matrix build/memory/etc. This must be set before initialising.

#         self._initialise()
#         self.initial_fill()
#         return self

#     def _initialise_submatrices(self):
#         from finesse.components import FrequencyGenerator

#         # Add in the diagonal elements of the matrix
#         # for n, node_inf in self._node_info.items():
#         #     Nsm = node_inf["nfreqs"]
#         #     Neq = node_inf["nhoms"]
#         #     for freq in range(Nsm):
#         #         fidx = self.findex(n, freq)  # Get index for this submatrix
#         #         comment = f"I,node={n.full_name},f={freq},fidx={fidx},Neq={Neq}"
#         #         self.M().add_diagonal_elements(Neq, fidx, comment)
#         #         Not building the diagonal currently

#         _done = {}
#         # Loop over every edge in the network which represents a bunch of
#         # connections between frequencies and HOMs between two nodes
#         for owner in self._edge_owners:
#             if owner in _done:
#                 continue

#             couples_f = isinstance(owner, FrequencyGenerator)

#             # For each connection this element wants...
#             for name in owner._registered_connections:
#                 # convert weak ref (input, output)
#                 nio = tuple(
#                     (owner.nodes[_] for _ in owner._registered_connections[name])
#                 )

#                 # If we are a carrier matrix only compute optics, no AC electronics or mechanics
#                 if not self.is_audio:
#                     if (
#                         nio[0].type is not NodeType.OPTICAL
#                         or nio[1].type is not NodeType.OPTICAL
#                     ):
#                         continue

#                 # Loop over all the frequencies we can couple between and add
#                 # submatrixes to the overall model
#                 for ifreq in self.frequencies:
#                     for ofreq in self.frequencies:
#                         # For each input and output frequency check if our
#                         # element wants to couple them at this
#                         if couples_f and not owner._couples_frequency(
#                             self, name, ifreq, ofreq
#                         ):
#                             continue
#                         elif not couples_f and ifreq != ofreq:
#                             # If it doesn't couple frequencies and the
#                             # frequencies are different then ignore
#                             continue

#                         iodx = []  # submatrix indices
#                         tags = []  # descriptive naming tags for submatrix key
#                         key_name = re.sub(r"^[^.]*\.", "", name)
#                         key_name = re.sub(r">[^.]*\.", ">", key_name)
#                         key = [owner, key_name]

#                         # Get simulation unique indices for submatrix
#                         # position. How we get these depends on the type of
#                         # the nodes involved
#                         for freq, node in zip((ifreq, ofreq), nio):
#                             if node.type is NodeType.OPTICAL:
#                                 iodx.append(self.findex(node, freq.index))
#                                 tags.append(freq.name)
#                                 key.append(freq)
#                             else:
#                                 # Mechanical and electrical don't have multiple
#                                 # freqs, so always use the zeroth frequency index
#                                 iodx.append(self.findex(node, 0))
#                                 tags.append("AC")
#                                 key.append(None)

#                         assert len(iodx) == 2
#                         assert len(key) == 4

#                         if tuple(key) not in self._submatrices:
#                             # smname = "{}__{}__{}".format(name, *tags)

#                             # print("Requesting:", *iodx, self.name, smname, i, j, nio)
#                             # Then we get a view of the underlying matrix which we set the values
#                             # with. Store one for each frequency. By requesting this view we are
#                             # telling the matrix that these elements should be non-zero in the
#                             # model.
#                             n1size = self._node_info[node]["nhoms"]
#                             n2size = self._node_info[node]["nhoms"]
#                             self._submatrices[tuple(key)] = np.zeros(
#                                 (n1size, n2size), dtype=complex
#                             )
#                         else:
#                             # Check if we've just requested the same submatrix.
#                             sm = self._submatrices[tuple(key)]
#                             if sm.from_idx != iodx[0] or sm.to_idx != iodx[1]:
#                                 raise Exception(
#                                     "Requested submatrix has already been requested,"
#                                     "but new one has different indices"
#                                 )
#                             else:
#                                 continue

#             _done[owner] = True

#     def __exit__(self, type_, value, traceback):
#         return

#     def _clear_rhs(self):
#         return None

#     def set_source(self, field_idx, vector):
#         return None



"""Library of coupling coefficient data structures and calculations.

Much of the code in this sub-package exists in a Cythonised form which
should only be used internally for calculating modal scattering during
simulations.

For the user, the :mod:`.tools` sub-module provides Python facing functions
for calculating certain types of coupling coefficients, and scattering
matrices of these, directly. The :mod:`.bayerhelms` module also contains
Python exposed functions for calculating Bayer-Helms :cite:`BayerHelms`
coefficients directly.

A data structure, :class:`.KnmMatrix`, representing a scattering matrix is
documented in :mod:`.matrix`. Objects of this type are returned whenever using
functions in this library which compute coupling coefficient matrices.

See :ref:`arbitrary_scatter_matrices` for details and examples on using some
of the Python-facing tools in this library.

Note that these tools use the same internal C code as that which gets executed
during simulations, and so will be very efficient off-the-shelf.
"""



"""Test parsing of cavity property detectors."""

import pytest
from finesse.script import parse
from finesse.detectors.cavity_detector import CP_KEYWORDS


@pytest.mark.parametrize("keyword,expected", CP_KEYWORDS.items())
def test_cavity_property_detector(keyword, expected):
    model = parse(
        f"""
        m m1 R=0.99 T=0.01
        m m2 R=0.99 T=0.01
        s s1 m1.p2 m2.p1
        cav cav1 m1.p2.o
        cp cav1cp cav1 {keyword}
        """
    )

    assert model.cav1cp.prop == expected



"""
Photodiode quantum noise detector.
"""

import numpy as np
from finesse.detectors.compute.quantum import (
    QND0Workspace,
    QNDNWorkspace,
    QShot0Workspace,
    QShotNWorkspace,
    QuantumNoiseDetectorWorkspace,
)
from finesse.detectors.general import Detector, NoiseDetector
from finesse.components.general import NoiseType
from finesse.parameter import float_parameter


class QuantumNoiseDetector(Detector, NoiseDetector):
    """
    Represents a quantum noise detector with no RF demodulations.

    It calculates the amplitude spectral density of the photocurrent
    noise of a photodiode output demodulated at the signal frequency.

    Parameters
    ----------
    name : str
        Name of newly created quantum noise detector.

    node : :class:`.Node`
        Node to read output from.

    nsr : bool, optional
        If true, calculate the noise-to-signal ratio rather than the absolute
        noise value.

    sources : list of :class:`.Connector`, optional
        If given, only detect quantum noise contributions from these components.

    exclude_sources : list of :class:`.Connector`, optional
        If given, don't detect quantum noise contributions from any of
        these components, even if given in `sources`.
    """

    def __init__(self, name, node, nsr=False, sources=None, exclude_sources=None):
        Detector.__init__(
            self, name, node, dtype=np.float64, unit="1/sqrt(Hz)", label="ASD"
        )
        NoiseDetector.__init__(self, NoiseType.QUANTUM)

        self.nsr = nsr
        self.sources = sources
        self.exclude_sources = exclude_sources

        self._request_selection_vector(name)

    def _has_sources(self):
        return self.sources is not None or self.exclude_sources is not None

    def _get_workspace(self, sims):
        ws = QND0Workspace(self, sims, self.nsr, self.sources, self.exclude_sources)
        return ws


@float_parameter("f", "Frequency")
@float_parameter("phase", "Phase")
class QuantumNoiseDetectorDemod1(Detector, NoiseDetector):
    """
    Represents a quantum noise detector with 1 RF demodulation.

    It calculates the amplitude spectral density of the photocurrent
    noise of a photodiode output demodulated at the signal frequency.

    Parameters
    ----------
    name : str
        Name of newly created quantum noise detector.

    node : :class:`.Node`
        Node to read output from.

    f : float
        Demodulation frequency in Hz

    phase : float
        Demodulation phase in degrees

    nsr : bool, optional
        If true, calculate the noise-to-signal ratio rather than the absolute
        noise value.

    sources : list of :class:`.Connector`, optional
        If given, only detect quantum noise contributions from these components.
    """

    def __init__(
        self, name, node, f, phase, nsr=False, sources=None, exclude_sources=None
    ):
        Detector.__init__(
            self, name, node, dtype=np.float64, unit="1/sqrt(Hz)", label="ASD"
        )
        NoiseDetector.__init__(self, NoiseType.QUANTUM)

        self.f = f
        self.phase = phase
        self.nsr = nsr
        self.sources = sources
        self.exclude_sources = exclude_sources

        self._request_selection_vector(name)

    def _has_sources(self):
        return self.sources is not None or self.exclude_sources is not None

    def _get_workspace(self, sims):
        ws = QNDNWorkspace(self, sims, 1, self.nsr, self.sources, self.exclude_sources)
        return ws


@float_parameter("f1", "Frequency")
@float_parameter("f2", "Frequency")
@float_parameter("phase1", "Phase")
@float_parameter("phase2", "Phase")
class QuantumNoiseDetectorDemod2(Detector, NoiseDetector):
    """
    Represents a quantum noise detector with 2 RF demodulations.

    It calculates the amplitude spectral density of the photocurrent
    noise of a photodiode output demodulated at the signal frequency.

    Parameters
    ----------
    name : str
        Name of newly created quantum noise detector.

    node : :class:`.Node`
        Node to read output from.

    f1 : float
        First demodulation frequency in Hz

    phase1 : float
        First demodulation phase in degrees

    f2 : float
        Second demodulation frequency in Hz

    phase2 : float
        Second demodulation phase in degrees

    nsr : bool, optional
        If true, calculate the noise-to-signal ratio rather than the absolute
        noise value.

    sources : list of :class:`.Connector`, optional
        If given, only detect quantum noise contributions from these components.
    """

    def __init__(
        self,
        name,
        node,
        f1,
        phase1,
        f2,
        phase2,
        nsr=False,
        sources=None,
        exclude_sources=None,
    ):
        Detector.__init__(
            self, name, node, dtype=np.float64, unit="1/sqrt(Hz)", label="ASD"
        )
        NoiseDetector.__init__(self, NoiseType.QUANTUM)

        self.f1 = f1
        self.f2 = f2
        self.phase1 = phase1
        self.phase2 = phase2
        self.nsr = nsr
        self.sources = sources
        self.exclude_sources = exclude_sources

        self._request_selection_vector(name)

    def _has_sources(self):
        return self.sources is not None or self.exclude_sources is not None

    def _get_workspace(self, sims):
        ws = QNDNWorkspace(self, sims, 2, self.nsr, self.sources, self.exclude_sources)
        return ws


class QuantumShotNoiseDetector(Detector):
    """
    Represents a quantum shot noise detector with no RF demodulations.

    It calculates the amplitude spectral density of the photocurrent
    noise of a photodiode output demodulated at the signal frequency,
    considering only vacuum noise contributions (ignoring radiation
    pressure and squeezing effects).

    Parameters
    ----------
    name : str
        Name of newly created quantum shot noise detector.

    node : :class:`.Node`
        Node to read output from.

    nsr : bool, optional
        If true, calculate the noise-to-signal ratio rather than the absolute
        noise value.
    """

    def __init__(self, name, node, nsr=False):
        self.nsr = nsr

        Detector.__init__(
            self, name, node, dtype=np.float64, unit="W/sqrt(Hz)", label="ASD"
        )

    def _get_workspace(self, sims):
        ws = QShot0Workspace(self, sims, self.nsr)
        return ws


@float_parameter("f", "Frequency")
@float_parameter("phase", "Phase")
class QuantumShotNoiseDetectorDemod1(Detector):
    """
    Represents a quantum shot noise detector with 1 RF demodulation.

    It calculates the amplitude spectral density of the photocurrent
    noise of a photodiode output demodulated at the signal frequency,
    considering only vacuum noise contributions (ignoring radiation
    pressure and squeezing effects).

    Parameters
    ----------
    name : str
        Name of newly created quantum shot noise detector.

    node : :class:`.Node`
        Node to read output from.

    f : float
        Demodulation frequency in Hz

    phase : float
        Demodulation phase in degrees

    nsr : bool, optional
        If true, calculate the noise-to-signal ratio rather than the absolute
        noise value.
    """

    def __init__(self, name, node, f, phase, nsr=False):
        Detector.__init__(
            self, name, node, dtype=np.float64, unit="W/sqrt(Hz)", label="ASD"
        )

        self.f = f
        self.phase = phase
        self.nsr = nsr

    def _get_workspace(self, sims):
        ws = QShotNWorkspace(self, sims, 1, self.nsr)
        return ws


@float_parameter("f1", "Frequency")
@float_parameter("f2", "Frequency")
@float_parameter("phase1", "Phase")
@float_parameter("phase2", "Phase")
class QuantumShotNoiseDetectorDemod2(Detector):
    """
    Represents a quantum shot noise detector with 2 RF demodulations.

    It calculates the amplitude spectral density of the photocurrent
    noise of a photodiode output demodulated at the signal frequency,
    considering only vacuum noise contributions (ignoring radiation
    pressure and squeezing effects).

    Parameters
    ----------
    name : str
        Name of newly created quantum shot noise detector.

    node : :class:`.Node`
        Node to read output from.

    f1 : float
        First demodulation frequency in Hz

    phase1 : float
        First demodulation phase in degrees

    f2 : float
        Second demodulation frequency in Hz

    phase2 : float
        Second demodulation phase in degrees

    nsr : bool, optional
        If true, calculate the noise-to-signal ratio rather than the absolute
        noise value.
    """

    def __init__(self, name, node, f1, phase1, f2, phase2, nsr=False):
        Detector.__init__(
            self, name, node, dtype=np.float64, unit="1/sqrt(Hz)", label="ASD"
        )

        self.f1 = f1
        self.f2 = f2
        self.phase1 = phase1
        self.phase2 = phase2
        self.nsr = nsr

    def _get_workspace(self, sims):
        ws = QShotNWorkspace(self, sims, 2, self.nsr)
        return ws


class GeneralQuantumNoiseDetector(Detector, NoiseDetector):
    """
    Represents a quantum noise detector.

    This detector calculates the amplitude spectral density of the
    photocurrent noise of a DC or demodulated photodiode output.

    Parameters
    ----------
    name : str
        Name of newly created quantum noise detector.

    node : :class:`.Node`
        Node to read output from.

    freqs : list of float or :class:`.Frequency`, optional
        List of mixer demodulation frequencies (in Hz).

    phases : list of float, optional
        List of mixer demodulation phases (in Hz).

    shot_only : bool, optional
        If True, detect only vacuum noise contributions.
    """

    def __init__(self, name, node, freqs=None, phases=None, shot_only=False):
        if freqs is None:
            freqs = []
        if phases is None:
            phases = []

        Detector.__init__(
            self, name, node, dtype=np.float64, unit="1/sqrt(Hz)", label="ASD"
        )
        NoiseDetector.__init__(self, NoiseType.QUANTUM)

        if len(freqs) == 0:
            self.__mode = "dc"
        elif len(phases) == len(freqs):
            self.__mode = "mixer"
        else:
            raise ValueError("'phases' must be as long as 'freqs'.")
        self.freqs = np.array(freqs)
        self.phases = np.array(phases)
        self.shot_only = shot_only
        self._request_selection_vector(name)

    def _get_workspace(self, sims):
        ws = QuantumNoiseDetectorWorkspace(self, sims)
        return ws



"""Tests of astigmatism at a node detection against analytics."""

import numpy as np
from numpy.testing import assert_allclose
import pytest
from finesse import Model
from finesse.analysis import xaxis
from finesse.utilities import changed_params


@pytest.fixture()
def simple_laser_model():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s0 L0.p1 BS.p1 L=1.2
    bs BS Rc=50
    s s1 BS.p2 END.p1 L=2.5
    nothing END

    gauss gL0 L0.p1.o w0=1m z=0

    astigd astig END.p1.i
    """
    )

    return IFO


@pytest.mark.parametrize(
    "scan",
    [
        ("gL0.w0x", 10e-6, 10e-2),
        ("gL0.w0y", 1.3e-5, 3e-3),
        ("gL0.zx", -5.4, 6.3),
        ("gL0.zy", -11.1, 2.7),
        ("BS.Rcx", 32.5, 76.2),
        ("BS.Rcy", -23.2, 90.1),
        ("BS.alpha", -12, 33),
    ],
)
def test_astig_detector(simple_laser_model: Model, scan):
    """Test detection of astigmatism at a node against analytics provided via
    AstigmaticPropagationSolution, for various parameter scans which induce astigmatic
    beams."""
    IFO = simple_laser_model

    sparam, lower, upper = scan
    param = IFO.reduce_get_attr(sparam)

    N = 100
    x = np.linspace(lower, upper, N)

    # Get the analytic form of the overlap between qx, qy
    # and evaluate it for the varied param array
    beam = IFO.propagate_beam_astig("L0.p1", "END.p1", symbolic=True)
    expect = 1 - beam.overlap(beam.end_node).eval(subs={param: x})

    # Can't scan alpha so do manual loop here instead
    if sparam == "BS.alpha":
        out = {"astig": np.zeros(N)}
        for i, a in enumerate(x):
            with changed_params({param: a}):
                out["astig"][i] = IFO.run()["astig"]

        # The above gives v. small astigmatism for small alpha so
        # set a suitable abs tolerance here
        atol = 1e-12
    else:
        out = xaxis(param, "lin", lower, upper, N - 1)
        atol = 0

    assert_allclose(out["astig"], expect, atol=atol)



import re
import pytest
from finesse.environment import show_tracebacks
from finesse.exceptions import FinesseException


@pytest.fixture
def finesse_exception():
    """A function that throws a :class:`.FinesseException`."""

    def excfunc():
        raise FinesseException("__test__")

    return excfunc


@pytest.mark.parametrize(
    "tracebacks,error",
    (
        (True, "\n__test__"),
        (False, ("\t(use finesse.tb() to see the full traceback)\n" "__test__")),
    ),
)
def test_show_traceback(finesse_exception, tracebacks, error):
    # NOTE: there is still a traceback shown in addition to the "use finesse.tb()"
    # message when Finesse is run from a non-Jupyter terminal and tracebacks are
    # switched off, because in such an environment the _render_traceback_ method of the
    # exception where the traceback is suppressed is ignored. This test is therefore not
    # technically testing the intended behaviour, but only a proxy.
    show_tracebacks(tracebacks)
    with pytest.raises(FinesseException, match=re.escape(error)):
        finesse_exception()



"""Finesse Sphinx extension.

This provides the `kat:command`, `kat:element` and `kat:analysis` directives and indices
for use in the documentation to document and cross-reference KatScript instructions.

Loosely based on https://www.sphinx-doc.org/en/master/development/tutorials/recipe.html
and https://github.com/click-contrib/sphinx-click/blob/master/sphinx_click/ext.py.

Author: Sean Leavey
"""

import abc
from functools import partial
from collections import defaultdict
from inspect import getabsfile
from docutils import nodes
from docutils.statemachine import string2lines, ViewList
from sphinx import addnodes
from sphinx.domains import Domain, Index
from sphinx.roles import XRefRole
from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import make_refnode, nested_parse_with_titles
from numpydoc.docscrape_sphinx import get_doc_object
from finesse.script.spec import KatSpec
from finesse.script.generator import KatSyntaxUnparser

__version__ = "0.7.0"

_SPEC = KatSpec()
_SYNTAXGEN = KatSyntaxUnparser()

_lines = partial(string2lines, tab_width=4, convert_whitespace=True)


def kat_syntax(adapter):
    """Build kat syntax string for `adapter`."""
    return _SYNTAXGEN.syntax(adapter.short_name)


class InstructionDirective(SphinxDirective, metaclass=abc.ABCMeta):
    """A custom Sphinx directive that describes a KatScript instruction."""

    description = "Instruction"
    has_content = True
    required_arguments = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        directive = self.arguments[0]
        try:
            self.adapter = _SPEC.directives[directive]
        except KeyError:
            raise ValueError(
                f"directive {directive} doesn't exist in KatScript language spec"
            )

    @property
    def doc_obj(self):
        """The numpydoc object."""
        return get_doc_object(self.adapter.docobj_type)

    def _description_slug(self):
        """Description type, usable as HTML attribute."""
        return nodes.fully_normalize_name(self.description)

    def _alias_slug(self, alias):
        return nodes.make_id(f"kat-{self._description_slug()}-{alias}")

    def _format_summary(self):
        yield from _lines(" ".join(self.doc_obj["Summary"]))
        yield ""

    def _format_syntax(self):
        yield f":Syntax: ``{kat_syntax(self.adapter)}``"
        yield ""

    def _format_param_text(self, name, params):
        yield f":{name}:"
        yield ""

        for pnames, pdoc in params:
            pnames = [f"``{pname.strip()}``" for pname in pnames.split(",")]
            # Indent the line so we get an inner definition list.
            line = f"  {', '.join(pnames)}"

            if pdoc:
                line += f": {pdoc}"

            yield line
            yield ""

        yield ""

    def run(self):
        """Main entry function, called by docutils upon encountering the directive.

        This generates the signature line (directive aliases), description, parameter
        lists and any other content specified in the restructuredText document.
        """
        try:
            source = getabsfile(self.adapter.call_signature_type)
        except TypeError:
            # This may be a Cythonised target. Use the string name instead.
            source = self.adapter.call_signature_type.__name__

        kat = self.env.get_domain("kat")

        if ":" not in self.name:
            raise ValueError(
                f"incorrect name format '{self.name}'; must be of the form 'kat:thing'"
            )

        self.domain, self.objtype = self.name.split(":", 1)

        # Index entry.
        indexnode = addnodes.index(entries=[])

        node = addnodes.desc()
        node.document = self.state.document
        node["domain"] = self.domain
        node["objtype"] = self.objtype
        node["noindex"] = noindex = "noindex" in self.options
        node["classes"].append(self.domain)
        node["classes"].append(node["objtype"])

        # Add each alias to the signature.
        signode = addnodes.desc_signature("", "", is_multiline=True)
        self.set_source_info(signode)
        node.append(signode)

        for i, alias in enumerate(self.adapter.aliases):
            is_primary = alias == self.adapter.full_name

            # All signatures have anchors due to the signode["ids"] line below, but
            # add_permalink shows a "#" next to the signature that users can copy.
            sigline = addnodes.desc_signature_line("", add_permalink=is_primary)
            sigline += addnodes.desc_name(text=alias)
            signode += sigline

            # Map index entry.
            if not noindex:
                anchor = self._alias_slug(alias)

                # Register an anchor for this alias.
                signode["ids"].append(anchor)

                kat.add_instruction(anchor, self, alias)

                if "noindexentry" not in self.options:
                    indextext = f"{alias} (KatScript {self.description})"
                    indexnode["entries"].append(("single", indextext, anchor, "", None))

        # The parameters.
        required_params = []
        optional_params = []

        for param_name, param_type, param_doc in self.doc_obj["Parameters"]:
            # Join together lines of pdoc.
            if param_doc:
                param_doc = " ".join(param_doc)
            else:
                param_doc = None

            param = param_name, param_doc

            if "optional" in param_type:
                optional_params.append(param)
            else:
                required_params.append(param)

        contentnode = addnodes.desc_content()
        self.set_source_info(contentnode)
        node.append(contentnode)

        content = ViewList()

        # The summary.
        for line in self._format_summary():
            content.append(line, source)
        content.append("", source)

        # The syntax.
        for line in self._format_syntax():
            content.append(line, source)
        content.append("", source)

        # Add parameters.
        if required_params:
            for line in self._format_param_text("Required", required_params):
                content.append(line, source)
        if optional_params:
            for line in self._format_param_text("Optional", optional_params):
                content.append(line, source)

        # Any extra content specified in the rST file.
        for line in self.content:
            content.append(line, source)
        content.append("", source)

        nested_parse_with_titles(self.state, content, contentnode)

        return [indexnode, node]


class CommandDirective(InstructionDirective):
    """A custom directive that describes a kat script command."""

    description = "Command"


class ElementDirective(InstructionDirective):
    """A custom directive that describes a kat script element."""

    description = "Element"


class AnalysisDirective(InstructionDirective):
    """A custom directive that describes a kat script analysis."""

    description = "Analysis"


class KatIndex(Index, metaclass=abc.ABCMeta):
    """A custom index that creates an instruction matrix."""

    index_type = None

    def generate(self, docnames=None):
        content = defaultdict(list)

        # Sort the list of instructions in alphabetical order.
        instructions = self.domain.get_objects()
        instructions = sorted(instructions, key=lambda instruction: instruction[0])

        # Generate the expected output, shown below, from the above using the first
        # letter of the recipe as a key to group thing.
        #
        # name, subtype, docname, anchor, extra, qualifier, description
        for name, dispname, typ, docname, anchor, _ in instructions:
            if typ != self.index_type:
                continue

            display_name = self.format_display_name(dispname)
            description = typ

            # The key is the index (in this case, the first character).
            content[display_name[0].lower()].append(
                (display_name, 0, docname, anchor, docname, "", description)
            )

        # Convert the dict to the sorted list of tuples expected.
        content = sorted(content.items())

        return content, True

    def format_display_name(self, name):
        return name


class CommandIndex(KatIndex):
    """A custom index that creates a command matrix."""

    name = "commandindex"
    localname = "Command Index"
    shortname = "Command"
    index_type = "Command"


class ElementIndex(KatIndex):
    """A custom index that creates an element matrix."""

    name = "elementindex"
    localname = "Element Index"
    shortname = "Element"
    index_type = "Element"


class AnalysisIndex(KatIndex):
    """A custom index that creates an analysis matrix."""

    name = "analysisindex"
    localname = "Analysis Index"
    shortname = "Analysis"
    index_type = "Analysis"


class KatDomain(Domain):
    name = "kat"
    label = "Kat Domain"
    roles = {
        "command": XRefRole(),
        "element": XRefRole(),
        "analysis": XRefRole(),
    }
    directives = {
        "command": CommandDirective,
        "element": ElementDirective,
        "analysis": AnalysisDirective,
    }
    indices = {
        CommandIndex,
        ElementIndex,
        AnalysisIndex,
    }
    initial_data = {
        "instructions": [],  # Object list.
    }

    def get_full_qualified_name(self, node):
        return "{}.{}".format("instruction", node.arguments[0])

    def get_objects(self):
        yield from self.data["instructions"]

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        # Match the specified reference to an alias of the available kat script
        # directives.
        if typ == "command":
            match = [
                (docname, anchor)
                for name, sig, otyp, docname, anchor, prio in self.get_objects()
                if (otyp == "Command" and target in _SPEC.commands[sig].aliases)
            ]
        elif typ == "element":
            match = [
                (docname, anchor)
                for name, sig, otyp, docname, anchor, prio in self.get_objects()
                if (otyp == "Element" and target in _SPEC.elements[sig].aliases)
            ]
        elif typ == "analysis":
            match = [
                (docname, anchor)
                for name, sig, otyp, docname, anchor, prio in self.get_objects()
                if (otyp == "Analysis" and target in _SPEC.analyses[sig].aliases)
            ]
        else:
            match = []

        if len(match) > 0:
            todocname = match[0][0]
            targ = match[0][1]

            return make_refnode(builder, fromdocname, todocname, targ, contnode, targ)
        else:
            print(f"Did not find kat script instruction for xref '{target}'")
            return None

    def add_instruction(self, anchor, instruction, signature):
        """Add a new instruction to the domain."""
        name = "{}.{}".format("instruction", signature)

        # name, dispname, type, docname, anchor, priority
        self.data["instructions"].append(
            (name, signature, instruction.description, self.env.docname, anchor, 0)
        )


def setup(app):
    app.add_domain(KatDomain)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import re
import pytest
from finesse.script.exceptions import KatScriptError
from finesse.symbols import OPERATORS, FUNCTIONS


@pytest.fixture
def compiler(compiler, fake_element_cls, fake_binop_cls, fake_unop_cls):
    # Define a command setter with fixed positional arg requirements so we can test incorrect
    # arguments.
    command = lambda a, b, c, d: None

    def set_command(model, *args, **kwargs):
        command(*args, **kwargs)

    compiler.spec.register_element(
        "fake_element", {"setter": fake_element_cls, "getter": fake_element_cls}
    )
    compiler.spec.register_command(
        "fake_command",
        {"setter": set_command, "getter": lambda: None, "singular": True,},
    )
    # Have to use real Finesse operators here because the builder matches against Finesse
    # operations.
    compiler.spec.binary_operators["+"] = OPERATORS["__add__"]
    compiler.spec.unary_operators["-"] = FUNCTIONS["neg"]

    return compiler


@pytest.mark.parametrize(
    "value", ("1", "-1", "1+1", "(1+1)", "'string'", "none", "inf", "-inf", "1.5+2e5j"),
)
def test_invalid_keyword_argument(compiler, value):
    """Test keyword arguments that don't exist in the Python constructor cause an error.

    Error handler uses the parsed value object so we have to check different types of
    value.
    """
    error = (
        f"\nline 1: 'fake_element' got an unexpected keyword argument 'c' (expected syntax: fake_element name a=none b=none)\n"
        f"-->1: fake_element myelement c={value}\n"
        f"                             ^"
    )
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(f"fake_element myelement c={value}")


@pytest.mark.parametrize(
    "script,error",
    (
        (
            "fake_command()",
            "\nline 1: 'fake_command' missing 4 required positional arguments: 'a', 'b', 'c', and 'd'\n"
            "-->1: fake_command()\n"
            "      ^^^^^^^^^^^^",
        ),
        (
            "fake_command(1)",
            "\nline 1: 'fake_command' missing 3 required positional arguments: 'b', 'c', and 'd'\n"
            "-->1: fake_command(1)\n"
            "      ^^^^^^^^^^^^",
        ),
        (
            "fake_command(1, 2)",
            "\nline 1: 'fake_command' missing 2 required positional arguments: 'c' and 'd'\n"
            "-->1: fake_command(1, 2)\n"
            "      ^^^^^^^^^^^^",
        ),
        (
            "fake_command(1, 2, 3)",
            "\nline 1: 'fake_command' missing 1 required positional argument: 'd'\n"
            "-->1: fake_command(1, 2, 3)\n"
            "      ^^^^^^^^^^^^",
        ),
    ),
)
def test_not_enough_positional_arguments(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,error",
    (
        (
            "fake_command(1, 2, 3, 4, 5)",
            "\nline 1: 'fake_command' takes 4 positional arguments but 5 were given\n"
            "-->1: fake_command(1, 2, 3, 4, 5)\n"
            "      ^^^^^^^^^^^^",
        ),
        (
            "fake_command(1, 2, 3, 4, 5, 6)",
            "\nline 1: 'fake_command' takes 4 positional arguments but 6 were given\n"
            "-->1: fake_command(1, 2, 3, 4, 5, 6)\n"
            "      ^^^^^^^^^^^^",
        ),
    ),
)
def test_too_many_positional_arguments(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,error",
    (
        (
            "fake_command(1, 2, 3, 4, 5, 6, a=1)",
            "\nline 1: 'fake_command' got multiple values for argument 'a'\n"
            "-->1: fake_command(1, 2, 3, 4, 5, 6, a=1)\n"
            "      ^^^^^^^^^^^^",
        ),
    ),
)
def test_duplicate_keyword_arguments(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)



import numpy as np

from finesse.detectors.general import Detector
from finesse.detectors.compute.gaussian import AstigmatismDetectorWorkspace


class AstigmatismDetector(Detector):
    """Detector for astigmatism figure-of-merit at a given node.

    The computed quantity is given via one minus :meth:`finesse.gaussian.BeamParam.overlap`.

    Parameters
    ----------
    name : str
        Name of the detector.

    node : :class:`.OpticalNode`
        Node to compute astigmatism at.
    """

    def __init__(self, name, node):
        super().__init__(name, node, dtype=np.float64, label="Astigmatism")

    @property
    def needs_fields(self):
        return False

    @property
    def needs_trace(self):
        return True

    def _get_workspace(self, sim):
        return AstigmatismDetectorWorkspace(self, sim)



from hypothesis import given, settings, note
from ......util import recursive_arrays


def stringify(array):
    if isinstance(array, list):
        return f"[{', '.join(stringify(item) for item in array)}]"
    return str(array)


@given(expected=recursive_arrays(operations=True))
@settings(deadline=1000)  # See #292.
def test_array_fuzzing(fuzz_value_parse_compare, expected):
    """Test that arrays parse properly."""
    array = stringify(expected)
    note(f"Array: {array}")
    fuzz_value_parse_compare(array, expected)
